import requests
from collections import defaultdict
from copy import deepcopy
import ssbio.utils
import os
import os.path as op
from lxml import etree
import logging
import glob

log = logging.getLogger(__name__)


def download_pisa_multimers_xml(pdb_ids, save_single_xml_files=True, outdir=None, force_rerun=False):
    """Download the PISA XML file for multimers.

    See: http://www.ebi.ac.uk/pdbe/pisa/pi_download.html for more info

    XML description of macromolecular assemblies:
        http://www.ebi.ac.uk/pdbe/pisa/cgi-bin/multimers.pisa?pdbcodelist
        where "pdbcodelist" is a comma-separated (strictly no spaces) list of PDB codes. The resulting file contain XML
        output of assembly data, equivalent to that displayed in PISA assembly pages, for each of the specified PDB
        entries.   NOTE: If a mass-download is intended, please minimize the number of retrievals by specifying as many
        PDB codes in the URL as feasible (20-50 is a good range), and never send another URL request until the previous
        one has been completed (meaning that the multimers.pisa file has been downloaded). Excessive requests will
        silently die in the server queue.

    Args:
        pdb_ids (str, list): PDB ID or list of IDs
        outdir (str): Directory to output PISA XML files
        force_rerun (bool): Redownload files if they already exist

    Returns:
        list: of files downloaded

    """
    if not outdir:
        outdir = os.getcwd()

    files = {}
    pdb_ids = ssbio.utils.force_lower_list(sorted(pdb_ids))

    # If we want to save single PISA XML files per PDB ID...
    if save_single_xml_files:
        # Check for existing PISA XML files
        if not force_rerun:
            existing_files = [op.basename(x) for x in glob.glob(op.join(outdir, '*_multimers.pisa.xml'))]
            # Store the paths to these files to return
            files = {v.split('_')[0]: op.join(outdir, v) for v in existing_files}
            log.debug('Already downloaded PISA files for {}'.format(list(files.keys())))
        else:
            existing_files = []

        # Filter PDB IDs based on existing file
        pdb_ids = [x for x in pdb_ids if '{}_multimers.pisa.xml'.format(x) not in existing_files]

        # Split the list into 50 to limit requests
        split_list = ssbio.utils.split_list_by_n(pdb_ids, 40)

        # Download PISA files
        for l in split_list:
            pdbs = ','.join(l)
            all_pisa_link = 'http://www.ebi.ac.uk/pdbe/pisa/cgi-bin/multimers.pisa?{}'.format(pdbs)
            r = requests.get(all_pisa_link)

            # Parse PISA file and save individual XML files
            parser = etree.XMLParser(ns_clean=True)
            tree = etree.fromstring(r.text, parser)
            for pdb in tree.findall('pdb_entry'):
                filename = op.join(outdir, '{}_multimers.pisa.xml'.format(pdb.find('pdb_code').text))
                add_root = etree.Element('pisa_multimers')
                add_root.append(pdb)
                with open(filename, 'wb') as f:
                    f.write(etree.tostring(add_root))
                files[pdb.find('pdb_code').text] = filename
                log.debug('{}: downloaded PISA results'.format(pdb))

    else:
        split_list = ssbio.utils.split_list_by_n(pdb_ids, 40)

        for l in split_list:
            pdbs = ','.join(l)
            all_pisa_link = 'http://www.ebi.ac.uk/pdbe/pisa/cgi-bin/multimers.pisa?{}'.format(pdbs)

            filename = op.join(outdir, '{}_multimers.pisa.xml'.format(pdbs))

            if ssbio.utils.force_rerun(flag=force_rerun, outfile=filename):
                r = requests.get(all_pisa_link)
                with open(filename, 'w') as f:
                    f.write(r.text)
                log.debug('Downloaded PISA results')
            else:
                log.debug('PISA results already downloaded')

            for x in l:
                files[x] = filename

    return files


def parse_pisa_multimers_xml(pisa_multimers_xml, download_structures=False, outdir=None, force_rerun=False):
    """Retrieve PISA information from an XML results file

    See: http://www.ebi.ac.uk/pdbe/pisa/pi_download.html for more info

    XML description of macromolecular assemblies:
        http://www.ebi.ac.uk/pdbe/pisa/cgi-bin/multimers.pisa?pdbcodelist
        where "pdbcodelist" is a comma-separated (strictly no spaces) list of PDB codes. The resulting file contain XML
        output of assembly data, equivalent to that displayed in PISA assembly pages, for each of the specified PDB
        entries.   NOTE: If a mass-download is intended, please minimize the number of retrievals by specifying as many
        PDB codes in the URL as feasible (20-50 is a good range), and never send another URL request until the previous
        one has been completed (meaning that the multimers.pisa file has been downloaded). Excessive requests will
        silently die in the server queue.

    Args:
        pdb_id: 4 character PDB ID

    Returns:

    """
    if not outdir:
        outdir = os.getcwd()

    parser = etree.XMLParser(ns_clean=True)
    tree = etree.parse(pisa_multimers_xml, parser)
    root = tree.getroot()

    pisa = defaultdict(dict)

    for pdb in root.findall('pdb_entry'):
        # Get the PDB ID
        pdb_id = pdb.find('pdb_code').text

        # Check the assembly status
        status = pdb.find('status').text
        errors = ['Entry not found', 'Overlapping structures', 'No symmetry operations']
        if status in errors:
            pisa[pdb_id]['status'] = status
            continue

        # Check monomer status
        num_complexes = int(pdb.find('total_asm').text)
        if num_complexes == 0:
            pisa[pdb_id]['status'] = 'MONOMER'
            continue

        elif num_complexes > 0:
            # All "assembly sets" (see PISA sets for more info)
            sets = pdb.findall('asm_set')

            for s in sets:
                set_id = int(s.find('ser_no').text)

                # All assemblies
                assemblies = s.findall('assembly')

                for cplx in assemblies:

                    ############################################################################################
                    # This part tells you the actual composition of the predicted complex (chains and ligands)
                    parts = cplx.findall('molecule')

                    chains = defaultdict(int)

                    for part in parts:
                        part_id = part.find('chain_id').text
                        if part_id.startswith('['):
                            part_id = 'LIG_' + part_id.split(']')[0].strip('[')
                        chains[str(part_id)] += 1

                    ligands = {}

                    for key in deepcopy(chains).keys():
                        if key.startswith('LIG_'):
                            ligands[str(key.split('_')[1])] = chains.pop(key)
                    ############################################################################################

                    adder = {}

                    cplx_id = int(cplx.find('id').text)
                    cplx_composition = str(cplx.find('composition').text)
                    d_g_diss = float(cplx.find('diss_energy').text)
                    d_g_int = float(cplx.find('int_energy').text)
                    pdb_biomol = int(cplx.find('R350').text)

                    if d_g_diss >= 0:
                        stable = True
                    else:
                        stable = False

                    adder['cplx_composition'] = cplx_composition.strip()
                    adder['cplx_chains'] = chains
                    adder['cplx_ligands'] = ligands
                    adder['stable'] = stable
                    adder['d_g_diss'] = d_g_diss
                    adder['d_g_int'] = d_g_int
                    adder['pdb_biomol'] = pdb_biomol

                    pisa[pdb_id][(set_id, cplx_id)] = adder

                    if download_structures:
                        ident = '{}:{},{}'.format(pdb_id, set_id, cplx_id)
                        filename = op.join(outdir, ident + '.pdb')

                        if ssbio.utils.force_rerun(flag=force_rerun, outfile=filename):
                            download_structure_link = 'http://www.ebi.ac.uk/pdbe/pisa/cgi-bin/multimer.pdb?{}'.format(
                                ident)
                            r = requests.get(download_structure_link)

                            with open(filename, 'w') as f:
                                f.write(r.text)

                            log.debug('{}: downloaded structure file'.format(ident))
                        else:
                            log.debug('{}: structure file already downloaded'.format(ident))

                        pisa[pdb_id][(set_id, cplx_id)]['structure_file'] = filename

    return pisa


def pdb_chain_stoichiometry_biomolone(pdbid):
    """Get the stoichiometry of the chains in biological assembly 1 as a dictionary.

    Steps taken are:
    1) Download PDB and parse header, make biomolecule if provided
    2) Count how many times each chain appears in biomolecule #1
    3) Convert chain id to uniprot id
    4) Return final dictionary

    Args:
        pdbid (str): 4 character PDB ID

    Returns:
        dict: {(ChainID,UniProtID): # occurences}
    """
    pass


import os.path as op
import unittest
from ssbio.pipeline.gempro import GEMPRO
from ssbio.pipeline.atlas import ATLAS


class TestATLAS(unittest.TestCase):
    """Unit tests for the ATLAS pipeline"""

    @classmethod
    def setUpClass(self):
        GEM_NAME = 'ecoli_test'
        ROOT_DIR = op.join('test_files', 'out')
        gem_file = op.join('test_files', 'Ec_core_flux1.xml')
        my_gempro = GEMPRO(GEM_NAME, ROOT_DIR, gem_file_path=gem_file, gem_file_type='sbml')

        # Prepare your ATLAS analysis
        self.atlas = ATLAS(atlas_name='atlas_test', base_gempro=my_gempro)



import logging
import os
import os.path as op
import sys

import cobra.flux_analysis
import cobra.manipulation
import numpy as np
import pandas as pd
from Bio import SeqIO
from cobra.core import DictList
from slugify import Slugify

import ssbio.cobra.utils
import ssbio.databases.ncbi
import ssbio.databases.patric
import ssbio.protein.sequence.properties.residues
import ssbio.protein.sequence.utils.alignment
import ssbio.protein.sequence.utils.blast
import ssbio.protein.sequence.utils.fasta
from ssbio import utils
from ssbio.core.object import Object
from ssbio.pipeline.gempro import GEMPRO

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
try:
    from IPython.display import clear_output
    have_ipython = True
    from tqdm import tqdm_notebook as tqdm
except ImportError:
    have_ipython = False
    from tqdm import tqdm

date = utils.Date()
custom_slugify = Slugify(safe_chars='-_.')
logging.basicConfig(stream=sys.stdout, level=logging.INFO)
log = logging.getLogger(__name__)


class ATLAS(Object):
    """Class to represent an ATLAS workflow to carry out multi-strain comparisons

    Main steps are:
    1. Strain-specific model construction based on orthologous genes & systems modeling
    2. Phylogenetic analysis to pick out important genes
    3. GEM-PRO of the "base strain"
    4. Structure property calculation & integrated structural systems analysis

    Each step may generate a report and also request additional files if something is missing
    """

    def __init__(self, atlas_name, base_gempro, base_genome_path=None, description=None):
        """Prepare a GEM-PRO model for ATLAS analysis

        Args:
            atlas_name (str): Name of your ATLAS project
            base_gempro (GEMPRO): Completed GEM-PRO model
            base_genome_path (str): Simple reference link to the genome FASTA file (CDS)
            description (str): Optional string to describe your project
        """
        Object.__init__(self, id=atlas_name, description=description)
        self.atlas_strains = DictList()

        # Load the GEM-PRO (could be a model, could just be a list of genes)
        self.base_strain_gempro = base_gempro

        # Check if there is a genome file associated with this model - if not, write all sequences and use that
        if not base_genome_path and not self.base_strain_gempro.genome_path:
            self.base_strain_gempro.genome_path = self.base_strain_gempro.write_representative_sequences_file(outname='BASE_CDS')
        else:
            self.base_strain_gempro.genome_path = base_genome_path
            # TODO: must check if base_genome gene IDs can be matched to the base model

        # Prepare ATLAS directories
        self.atlas_dir = None
        self.atlas_model_dir = None
        self.atlas_data_dir = None
        self.atlas_sequence_dir = None
        self.atlas_seq_genomes_dir = None
        self.atlas_seq_genes_dir = None
        self._create_dirs(self.base_strain_gempro.base_dir)

        # Other attributes
        self.df_orthology_matrix = pd.DataFrame()
        # Mark if the orthology matrix has gene IDs (thus we need to retrieve seqs from the genome file) or if
        # it is in the orthology matrix itself
        self._orthology_matrix_has_sequences = False

    def _create_dirs(self, root_dir):
        """Internal method to create ATLAS directories.

        Args:
            root_dir (str): Root directory to create a single new ATLAS project folder in

        """
        list_of_dirs = []

        if not root_dir:
            root_dir = os.getcwd()

        if not op.exists(root_dir):
            raise ValueError('{}: folder does not exist'.format(root_dir))

        # atlas_dir - directory where all ATLAS analysis will be carried out
        self.atlas_dir = op.join(root_dir, 'atlas')
        list_of_dirs.append(self.atlas_dir)

        # atlas_model_dir - directory where new strain specific models will be stored
        self.atlas_model_dir = op.join(self.atlas_dir, 'models')
        list_of_dirs.append(self.atlas_model_dir)

        # atlas_data_dir - directory where all data will be stored
        self.atlas_data_dir = op.join(self.atlas_dir, 'data')
        list_of_dirs.append(self.atlas_data_dir)

        # atlas_sequence_dir - sequence related files are stored here
        self.atlas_sequence_dir = op.join(self.atlas_dir, 'sequences')
        self.atlas_seq_genomes_dir = op.join(self.atlas_sequence_dir, 'by_organism')
        self.atlas_seq_genes_dir = op.join(self.atlas_sequence_dir, 'by_gene')
        list_of_dirs.extend([self.atlas_sequence_dir, self.atlas_seq_genomes_dir, self.atlas_seq_genes_dir])

        # Make the dirs
        for directory in list_of_dirs:
            if not op.exists(directory):
                os.makedirs(directory)
                log.info('Created directory: {}'.format(directory))
            else:
                log.debug('Directory already exists: {}'.format(directory))

        log.info('{}: ATLAS project files location'.format(self.atlas_dir))

    def load_manual_orthology_matrix(self, df, clean_names=True,
                                     remove_strains_with_no_orthology=True,
                                     remove_strains_with_no_differences=False,
                                     remove_genes_not_in_base_model=True):
        """Load a manually curated orthology matrix to use in ATLAS. Genes = rows, strains = columns.

        Args:
            df (DataFrame): Pandas DataFrame with genes as the rows and strains as the columns
            clean_names (bool): Remove unwanted characters from gene names and strain IDs
            remove_strains_with_no_orthology (bool): Remove strains which have no orthologous genes found
            remove_strains_with_no_differences (bool): Remove strains which have all the same genes as the base model.
                Default is False because since orthology is found using a PID cutoff, all genes may be present but
                differences may be on the sequence level.
            remove_genes_not_in_base_model (bool): Remove genes from the orthology matrix which are not present in our
                base model. This happens if we use a genome file for our model that has other genes in it.

        """
        self._orthology_matrix_has_sequences = True

        if clean_names:
            new_rows = [custom_slugify(x) for x in df.index]
            new_cols = [custom_slugify(y) for y in df.columns]
            df.index = new_rows
            df.columns = new_cols

        self.df_orthology_matrix = df

        # Make the copies of the base model
        for strain_id in tqdm(self.df_orthology_matrix.columns):
            self._copy_base_gempro(new_id=strain_id)

        # Filter the strains and orthology matrix
        self._filter_with_orthology_matrix(remove_strains_with_no_orthology=remove_strains_with_no_orthology,
                                           remove_strains_with_no_differences=remove_strains_with_no_differences,
                                           remove_genes_not_in_base_model=remove_genes_not_in_base_model)

    def _copy_base_gempro(self, new_id):
        """Copy the base strain GEM-PRO into a new GEM-PRO with a specified ID.

        Appends the model to the atlas_strains attribute.

        Args:
            new_id (str): New ID to be assigned to the copied model

        Returns:
            GEMPRO: copied GEM-PRO to represent the new strain

        """
        # Temporarily disable logging messages
        logging.disable(logging.WARNING)
        if self.base_strain_gempro.model:
            # If there is a SBML model associated with the GEMPRO, copy that model
            copied_model = GEMPRO(gem_name=new_id, create_dirs=False, gem=self.base_strain_gempro.model.copy())
            copied_model.model.id = new_id

            # Reset the GenePro attributes
            for x in copied_model.genes:
                x.reset_protein()
        else:
            # Otherwise, just copy the list of genes over and rename the IDs
            strain_genes = [x.id for x in self.base_strain_gempro.genes]
            copied_model = GEMPRO(gem_name=new_id, create_dirs=False, genes_list=strain_genes)
        # Re-enable logging
        logging.disable(logging.NOTSET)

        self.atlas_strains.append(copied_model)
        log.debug('{}: new model copied from base model'.format(new_id))

        return self.atlas_strains.get_by_id(new_id)

    def load_strain_ids_and_genomes(self, ids_to_genome_file):
        """Load a dictionary of PATRIC IDs which point to the path of their respective genome files.

        Creates initial copies of the base strain model for each strain ID and stores them in self.atlas_strains

        Args:
            ids_to_genome_file (dict): Keys are PATRIC IDs (which will become your strain IDs) and
                values are absolute paths to the FASTA file containing CDS regions

        """
        for strain_id, genome_path in tqdm(ids_to_genome_file.items()):
            strain_gp = self._copy_base_gempro(new_id=strain_id)
            strain_gp.genome_path = genome_path

    def download_genome_cds_patric(self, ids, force_rerun=False):
        """Download genome files from PATRIC give a list of IDs

        Args:
            ids (str, list): PATRIC ID or list of PATRIC IDs
            force_rerun: If genome files should be downloaded again even if they exist

        """
        strains_to_fasta_file = {}
        ids = ssbio.utils.force_list(ids)

        log.info('Downloading sequences from PATRIC...')
        for patric_id in tqdm(ids):
            f = ssbio.databases.patric.download_coding_sequences(patric_id=patric_id, seqtype='protein',
                                                                 outdir=self.atlas_seq_genomes_dir,
                                                                 force_rerun=force_rerun)
            if f:
                strains_to_fasta_file[patric_id] = f
                log.debug('{}: downloaded sequence'.format(patric_id))
            else:
                log.warning('{}: unable to download sequence'.format(patric_id))

        log.info('Copying base strain to create initial strain specific models...')
        self.load_strain_ids_and_genomes(strains_to_fasta_file)
        log.info('Created {} new strain GEM-PROs, accessible at "atlas_strains"'.format(len(strains_to_fasta_file)))

    def get_orthology_matrix(self, pid_cutoff=None, bitscore_cutoff=None, evalue_cutoff=None, filter_condition='OR',
                             remove_strains_with_no_orthology=True,
                             remove_strains_with_no_differences=False,
                             remove_genes_not_in_base_model=True):
        """Create the orthology matrix by finding best bidirectional BLAST hits. Genes = rows, strains = columns

        Runs run_makeblastdb, run_bidirectional_blast, and calculate_bbh for protein sequences.

        Args:
            pid_cutoff (float): Minimum percent identity between BLAST hits to filter for in the range [0, 100]
            bitscore_cutoff (float): Minimum bitscore allowed between BLAST hits
            evalue_cutoff (float): Maximum E-value allowed between BLAST hits
            filter_condition (str): 'OR' or 'AND', how to combine cutoff filters. 'OR' gives more results since it
            is less stringent, as you will be filtering for hits with (>80% PID or >30 bitscore or <0.0001 evalue).
            remove_strains_with_no_orthology (bool): Remove strains which have no orthologous genes found
            remove_strains_with_no_differences (bool): Remove strains which have all the same genes as the base model.
                Default is False because since orthology is found using a PID cutoff, all genes may be present but
                differences may be on the sequence level.
            remove_genes_not_in_base_model (bool): Remove genes from the orthology matrix which are not present in our
                base model. This happens if we use a genome file for our model that has other genes in it.

        Returns:
            DataFrame: Orthology matrix calculated from best bidirectional BLAST hits.

        """
        # TODO: document and test other cutoffs

        # Get the path to the reference genome
        r_file = self.base_strain_gempro.genome_path

        bbh_files = {}

        for strain_model in tqdm(self.atlas_strains):
            g_file = strain_model.genome_path

            # Run bidirectional BLAST
            log.debug('{} vs {}: Running bidirectional BLAST'.format(self.base_strain_gempro.id, strain_model.id))
            r_vs_g, g_vs_r = ssbio.protein.sequence.utils.blast.run_bidirectional_blast(reference=r_file, other_genome=g_file,
                                                                                        dbtype='prot',
                                                                                        outdir=self.atlas_seq_genomes_dir)

            # Using the BLAST files, find the BBH
            log.debug('{} vs {}: Finding BBHs'.format(self.base_strain_gempro.id, strain_model.id))
            bbh = ssbio.protein.sequence.utils.blast.calculate_bbh(blast_results_1=r_vs_g, blast_results_2=g_vs_r,
                                                                   outdir=self.atlas_seq_genomes_dir)
            bbh_files[strain_model.id] = bbh

        # Make the orthologous genes matrix
        log.debug('Creating orthology matrix')
        ortho_matrix = ssbio.protein.sequence.utils.blast.create_orthology_matrix(r_name=self.base_strain_gempro.id,
                                                                                  genome_to_bbh_files=bbh_files,
                                                                                  pid_cutoff=pid_cutoff,
                                                                                  bitscore_cutoff=bitscore_cutoff,
                                                                                  evalue_cutoff=evalue_cutoff,
                                                                                  filter_condition=filter_condition,
                                                                                  outname='{}_{}_orthology.csv'.format(self.base_strain_gempro.id, 'prot'),
                                                                                  outdir=self.atlas_data_dir)

        log.info('Saved orthology matrix at {}. See the "df_orthology_matrix" attribute.'.format(ortho_matrix))
        self.df_orthology_matrix = pd.read_csv(ortho_matrix, index_col=0)

        # Filter the matrix to genes only in our analysis, and also check for strains with no differences or no orthologous genes
        self._filter_with_orthology_matrix(remove_strains_with_no_orthology=remove_strains_with_no_orthology,
                                           remove_strains_with_no_differences=remove_strains_with_no_differences,
                                           remove_genes_not_in_base_model=remove_genes_not_in_base_model)

        return self.df_orthology_matrix

    def _filter_with_orthology_matrix(self,
                                      remove_strains_with_no_orthology=True,
                                      remove_strains_with_no_differences=False,
                                      remove_genes_not_in_base_model=True):
        """Filters the orthology matrix by removing genes not in our base model, and also
            removes strains from the analysis which have: 0 orthologous genes or no difference from the base strain.

        Args:
            remove_strains_with_no_orthology (bool): Remove strains which have no orthologous genes found
            remove_strains_with_no_differences (bool): Remove strains which have all the same genes as the base model.
                Default is False because since orthology is found using a PID cutoff, all genes may be present but
                differences may be on the sequence level.
            remove_genes_not_in_base_model (bool): Remove genes from the orthology matrix which are not present in our
                base model. This happens if we use a genome file for our model that has other genes in it.

        """

        if len(self.df_orthology_matrix) == 0:
            raise RuntimeError('Empty orthology matrix')

        initial_num_strains = len(self.atlas_strains)

        # Adding names to the row and column of the orthology matrix
        self.df_orthology_matrix = self.df_orthology_matrix.rename_axis('gene').rename_axis("strain", axis="columns")

        # Gene filtering (of the orthology matrix)
        if remove_genes_not_in_base_model:
            # Check for gene IDs that are in the model and not in the orthology matrix
            # This is probably because: the CDS FASTA file for the base strain did not contain the correct ID
            # for the gene and consequently was not included in the orthology matrix
            # Save these and report them
            base_strain_gene_ids = [x.id for x in self.base_strain_gempro.genes]
            self.missing_in_orthology_matrix = [x for x in base_strain_gene_ids if x not in self.df_orthology_matrix.index.tolist()]
            self.missing_in_base_strain = [y for y in self.df_orthology_matrix.index.tolist() if y not in base_strain_gene_ids]

            # Filter the matrix for genes within our base model only
            self.df_orthology_matrix = self.df_orthology_matrix[self.df_orthology_matrix.index.map(lambda x: x in base_strain_gene_ids)]
            log.info('Filtered orthology matrix for genes present in base model')
            log.warning('{} genes are in your base model but not your orthology matrix, see the attribute "missing_in_orthology_matrix"'.format(len(self.missing_in_orthology_matrix)))
            log.warning('{} genes are in the orthology matrix but not your base model, see the attribute "missing_in_base_strain"'.format(len(self.missing_in_base_strain)))

        # Strain filtering
        for strain_model in self.atlas_strains.copy():
            if remove_strains_with_no_orthology:
                if strain_model.id not in self.df_orthology_matrix.columns:
                    self.atlas_strains.remove(strain_model)
                    log.info('{}: no orthologous genes found for this strain, removed from analysis.'.format(strain_model.id))
                    continue
                elif self.df_orthology_matrix[strain_model.id].isnull().all():
                    self.atlas_strains.remove(strain_model)
                    log.info('{}: no orthologous genes found for this strain, removed from analysis.'.format(strain_model.id))
                    continue

            if remove_strains_with_no_differences:
                not_in_strain = self.df_orthology_matrix[pd.isnull(self.df_orthology_matrix[strain_model.id])][strain_model.id].index.tolist()
                if len(not_in_strain) == 0:
                    self.atlas_strains.remove(strain_model)
                    log.info('{}: strain has no differences from the base, removed from analysis.')
                    continue

        log.info('{} strains to be analyzed, {} strains removed'.format(len(self.atlas_strains), initial_num_strains-len(self.atlas_strains)))

    def _pare_down_model(self, gempro, genes_to_remove):
        """Mark genes as non-functional in a GEM-PRO. If there is a COBRApy model associated with it, the
            COBRApy method delete_model_genes is utilized to delete genes.

        Args:
            gempro (GEMPRO): GEMPRO object
            genes_to_remove (list): List of gene IDs to remove from the model

        """
        # Filter out genes in genes_to_remove which do not show up in the model
        model_genes = [x.id for x in gempro.genes]
        genes_to_remove = list(set(genes_to_remove).intersection(set(model_genes)))
        genes_to_remove.extend(self.missing_in_orthology_matrix)

        if len(genes_to_remove) == 0:
            log.info('{}: no genes marked for removal'.format(gempro.id))
            return
        else:
            log.debug('{}: {} genes marked for removal'.format(gempro.id, len(genes_to_remove)))

        # If a COBRApy model exists, utilize the delete_model_genes method
        if gempro.model:
            gempro.model._trimmed = False
            gempro.model._trimmed_genes = []
            gempro.model._trimmed_reactions = {}

            # Delete genes!
            cobra.manipulation.delete_model_genes(gempro.model, genes_to_remove)

            if gempro.model._trimmed:
                log.info('{}: deleted {} reactions, {} genes'.format(gempro.model.id,
                                                                     len(gempro.model._trimmed_reactions),
                                                                     len(gempro.model._trimmed_genes)))
        # Otherwise, just mark the genes as non-functional
        else:
            for g in genes_to_remove:
                my_gene = gempro.genes.get_by_id(g)
                my_gene.functional = False
            log.info('{}: deleted {} genes'.format(gempro.id, len(genes_to_remove)))

    def build_strain_specific_models(self):
        """Using the orthologous genes matrix, modify the strain specific models based on if orthologous genes exist.

        Also stores the sequences directly in the GEM-PRO protein sequences attribute for the strains.
        """

        if len(self.df_orthology_matrix) == 0:
            raise RuntimeError('Empty orthology matrix')

        # For each genome, create the strain specific model
        for strain_model in tqdm(self.atlas_strains):
            log.debug('{}: building strain specific model'.format(strain_model.id))

            # Get a list of genes which do not have orthology in the strain
            not_in_strain = self.df_orthology_matrix[pd.isnull(self.df_orthology_matrix[strain_model.id])][strain_model.id].index.tolist()

            # Mark genes non-functional
            self._pare_down_model(gempro=strain_model, genes_to_remove=not_in_strain)

            # Load sequences into the base and strain models
            self._load_strain_sequences(strain_model.id)

        log.info('Created {} new strain-specific models and loaded in sequences'.format(len(self.atlas_strains)))

    def _load_strain_sequences(self, strain_id):
        """Load strain sequences from the orthology matrix into the base model for comparisons, and into the
            strain-specific model itself.

        """
        strain_model = self.atlas_strains.get_by_id(strain_id)

        if self._orthology_matrix_has_sequences:  # Load directly from the orthology matrix if it contains sequences
            strain_sequences = self.df_orthology_matrix[strain_id].to_dict()
        else:  # Otherwise load from the genome file if the orthology matrix contains gene IDs
            # Load the genome FASTA file
            log.debug('{}: loading strain genome CDS file'.format(strain_model.genome_path))
            strain_sequences = SeqIO.index(strain_model.genome_path, 'fasta')

        for strain_gene in strain_model.genes:
            if strain_gene.functional:
                if self._orthology_matrix_has_sequences:
                    strain_gene_key = strain_gene.id
                else:
                    # Pull the gene ID of the strain from the orthology matrix
                    strain_gene_key = self.df_orthology_matrix.loc[strain_gene.id, strain_model.id]
                    log.debug('{}: original gene ID in strain fasta file'.format(strain_gene_key))

                # Load into the base strain for comparisons
                base_gene = self.base_strain_gempro.genes.get_by_id(strain_gene.id)
                new_id = '{}_{}'.format(strain_gene.id, strain_id)
                if base_gene.protein.sequences.has_id(new_id):
                    log.debug('{}: sequence already loaded into base model'.format(new_id))
                    continue
                base_gene.protein.load_manual_sequence(seq=strain_sequences[strain_gene_key], ident=new_id,
                                                       set_as_representative=False)
                log.debug('{}: loaded sequence into base model'.format(new_id))

                # Load into the strain GEM-PRO
                strain_gene.protein.load_manual_sequence(seq=strain_sequences[strain_gene_key], ident=new_id,
                                                         set_as_representative=True)
                log.debug('{}: loaded sequence into strain model'.format(new_id))

    def align_orthologous_genes_pairwise(self, gapopen=10, gapextend=0.5):
        """For each gene in the base strain, run a pairwise alignment for all orthologous gene sequences to it."""
        for base_gene in tqdm(self.base_strain_gempro.genes):
            if len(base_gene.protein.sequences) > 1:
                alignment_dir = op.join(self.atlas_seq_genes_dir, base_gene.id)
                if not op.exists(alignment_dir):
                    os.mkdir(alignment_dir)
                base_gene.protein.pairwise_align_sequences_to_representative(gapopen=gapopen, gapextend=gapextend,
                                                                             outdir=alignment_dir, parse=True)

    def align_orthologous_genes_multiple(self):
        """For each gene in the base strain, run a multiple alignment to all orthologous strain genes"""
        pass

    def get_atlas_summary_df(self):
        """Create a single data frame which summarizes all genes per row.

        Returns:
            DataFrame: Pandas DataFrame of the results

        """
        all_info = []
        for g in self.base_strain_gempro.genes_with_a_representative_sequence:
            info = {}
            info['Gene_ID'] = g.id

            # Protein object
            p = g.protein
            info['Protein_sequences'] = len(p.sequences)
            info['Protein_structures'] = len(p.structures)

            # SeqProp
            rseq = p.representative_sequence
            if rseq.seq_record:
                info['RepSeq_sequence_length'] = rseq.seq_len
                info['RepSeq_num_sequence_alignments'] = len(rseq.sequence_alignments)
                info['RepSeq_num_structure_alignments'] = len(rseq.structure_alignments)

                # SeqRecord annotations (properties calculated that summarize the whole sequence)
                rseq_sr = rseq.seq_record
                for annotation_name, annotation in rseq_sr.annotations.items():
                    info['RepSeq_' + annotation_name] = annotation

                # SeqRecord alignment annotations
                all_num_mutations = []
                all_num_deletions = []
                all_len_deletions = []
                all_num_insertions = []
                all_len_insertions = []
                all_percent_identity = []
                all_percent_similarity = []
                for aln in rseq.sequence_alignments:
                    # Gather the strain speicific stuff
                    if '{}_'.format(p.id) not in aln.annotations['b_seq']:
                        continue
                    info[aln.annotations['b_seq'].split('{}_'.format(p.id))[1]] = aln.annotations['percent_identity']

                    # Gather the percent identities/similarities
                    all_percent_identity.append(aln.annotations['percent_identity'])
                    all_percent_similarity.append(aln.annotations['percent_similarity'])

                    # Gather the number of residues that are mutated (filter for different mutations of same residue)
                    num_mutations = len(list(set([x[1] for x in aln.annotations['mutations']])))
                    all_num_mutations.append(num_mutations)

                    # Gather the number of deletions as well as the length of the deletion
                    if not aln.annotations['deletions']:
                        num_deletions = 0
                        len_deletions = [0]
                    else:
                        num_deletions = len(aln.annotations['deletions'])
                        len_deletions = [x[1] for x in aln.annotations['deletions']]
                    all_num_deletions.append(num_deletions)
                    # Get the total length of the deletion for this one strain
                    avg_len_deletions = np.sum(len_deletions)
                    all_len_deletions.append(avg_len_deletions)

                    # Gather the number of insertions as well as the length of the insertion
                    if not aln.annotations['insertions']:
                        num_insertions = 0
                        len_insertions = [0]
                    else:
                        num_insertions = len(aln.annotations['insertions'])
                        len_insertions = [x[1] for x in aln.annotations['insertions']]
                    all_num_insertions.append(num_insertions)
                    # Get the total length of insertion for this one strain
                    avg_len_insertions = np.sum(len_insertions)
                    all_len_insertions.append(avg_len_insertions)

                info['ATLAS_mean_num_mutations'] = np.mean(all_num_mutations)
                info['ATLAS_mean_num_deletions'] = np.mean(all_num_deletions)
                info['ATLAS_mean_len_deletions'] = np.mean(all_len_deletions)
                info['ATLAS_mean_num_insertions'] = np.mean(all_num_insertions)
                info['ATLAS_mean_len_insertions'] = np.mean(all_len_insertions)
                info['ATLAS_mean_percent_identity'] = np.mean(all_percent_identity)
                info['ATLAS_mean_percent_similarity'] = np.mean(all_percent_similarity)

                # Other mutation analysis
                single, fingerprint = rseq.sequence_mutation_summary()

                # Mutations that show up in more than 10% of strains
                singles = []
                for k, v in single.items():
                    k = [str(x) for x in k]
                    if len(v) / len(rseq.sequence_alignments) >= 0.01:
                        singles.append(''.join(k))  # len(v) is the number of strains which have this mutation
                info['ATLAS_popular_mutations'] = ';'.join(singles)

                # Mutation groups that show up in more than 10% of strains
                allfingerprints = []
                for k, v in fingerprint.items():
                    if len(v) / len(rseq.sequence_alignments) >= 0.01:
                        fingerprints = []
                        for m in k:
                            y = [str(x) for x in m]
                            fingerprints.append(''.join(y))
                        allfingerprints.append('-'.join(fingerprints))
                info['ATLAS_popular_mutation_groups'] = ';'.join(allfingerprints)

            # StructProp
            rstruct = p.representative_structure
            if rstruct:
                if rstruct.structure_file:
                    info['RepStruct_ID'] = rstruct.id
                    info['RepStruct_is_experimental'] = rstruct.is_experimental
                    info['RepStruct_description'] = rstruct.description
                    info['RepStruct_repseq_coverage'] = rstruct.reference_seq_top_coverage

                    # ChainProp
                    rchain = rstruct.representative_chain
                    info['RepChain_ID'] = rchain.id

                    # ChainProp SeqRecord annotations
                    rchain_sr = rchain.seq_record
                    for annotation_name, annotation in rchain_sr.annotations.items():
                        info['RepChain_' + annotation_name] = annotation

            all_info.append(info)

        cols = ['Gene_ID', 'Protein_sequences', 'Protein_structures',
                'RepSeq_num_sequence_alignments', 'RepSeq_num_structure_alignments', 'RepSeq_num_tm_helix-tmhmm',
                'RepSeq_sequence_length',
                'RepStruct_ID', 'RepStruct_description', 'RepStruct_is_experimental', 'RepStruct_repseq_coverage',
                'RepChain_ID',
                'ATLAS_mean_percent_identity', 'ATLAS_mean_percent_similarity', 'ATLAS_mean_num_mutations',
                'ATLAS_popular_mutations', 'ATLAS_popular_mutation_groups', 'ATLAS_mean_num_deletions',
                'ATLAS_mean_num_insertions', 'ATLAS_mean_len_deletions', 'ATLAS_mean_len_insertions',
                'RepSeq_aromaticity', 'RepSeq_instability_index', 'RepSeq_isoelectric_point', 'RepSeq_molecular_weight',
                'RepSeq_monoisotopic', 'RepSeq_percent_acidic', 'RepSeq_percent_aliphatic', 'RepSeq_percent_aromatic',
                'RepSeq_percent_B-sspro8', 'RepSeq_percent_basic', 'RepSeq_percent_buried-accpro',
                'RepSeq_percent_buried-accpro20', 'RepSeq_percent_C-sspro', 'RepSeq_percent_C-sspro8',
                'RepSeq_percent_charged', 'RepSeq_percent_E-sspro', 'RepSeq_percent_E-sspro8',
                'RepSeq_percent_exposed-accpro', 'RepSeq_percent_exposed-accpro20', 'RepSeq_percent_G-sspro8',
                'RepSeq_percent_H-sspro', 'RepSeq_percent_H-sspro8', 'RepSeq_percent_helix_naive',
                'RepSeq_percent_I-sspro8', 'RepSeq_percent_non-polar', 'RepSeq_percent_polar',
                'RepSeq_percent_S-sspro8', 'RepSeq_percent_small', 'RepSeq_percent_strand_naive',
                'RepSeq_percent_T-sspro8', 'RepSeq_percent_tiny', 'RepSeq_percent_turn_naive',
                'RepChain_percent_B-dssp', 'RepChain_percent_C-dssp', 'RepChain_percent_E-dssp',
                'RepChain_percent_G-dssp', 'RepChain_percent_H-dssp', 'RepChain_percent_I-dssp',
                'RepChain_percent_S-dssp', 'RepChain_percent_T-dssp', 'RepChain_SSBOND-biopython']
        cols.extend([x.id for x in self.atlas_strains])
        df_atlas_summary = pd.DataFrame(all_info, columns=cols)
        # Drop columns that don't have anything in them
        df_atlas_summary.dropna(axis=1, how='all', inplace=True)

        return df_atlas_summary

    def get_atlas_per_gene_mutation_df(self, gene_id):
        """Create a single data frame which summarizes a gene and its mutations.

        Args:
            gene_id (str): Gene ID in the base model

        Returns:
            DataFrame: Pandas DataFrame of the results

        """
        # TODO: also count: number of unique mutations (have to consider position, amino acid change)
        # TODO: keep track of strain with most mutations, least mutations
        # TODO: keep track of strains that conserve the length of the protein, others that extend or truncate it
        # need statistical test for that too (how long is "extended"/"truncated"?)
        # TODO: number of strains with at least 1 mutations
        # TODO: number of strains with <5% mutated, 5-10%, etc

        g = self.base_strain_gempro.genes.get_by_id(gene_id)

        single, fingerprint = g.protein.representative_sequence.sequence_mutation_summary()

        structure_type_suffix = 'NA'
        appender = []

        for k, strains in single.items():

            # Mutations in the strain
            to_append = {}
            orig_res = k[0]
            resnum = int(k[1])
            mutated_res = k[2]
            num_strains = len(strains)
            strain_ids = [str(x.split(g.id + '_')[1]) for x in strains]
            to_append['base_residue'] = orig_res
            to_append['base_resnum'] = resnum
            to_append['strain_residue'] = mutated_res
            to_append['num_strains'] = num_strains
            to_append['strains_mutated'] = ';'.join(strain_ids)
            to_append['at_disulfide_bridge'] = False

            # Residue properties
            origres_props = ssbio.protein.sequence.properties.residues.residue_biochemical_definition(orig_res)
            mutres_props = ssbio.protein.sequence.properties.residues.residue_biochemical_definition(mutated_res)
            to_append['base_residue_prop'] = origres_props
            to_append['strain_residue_prop'] = mutres_props

            # Grantham score - score a mutation based on biochemical properties
            grantham_s, grantham_txt = ssbio.protein.sequence.properties.residues.grantham_score(orig_res, mutated_res)
            to_append['grantham_score'] = grantham_s
            to_append['grantham_annotation'] = grantham_txt

            # Structure properties - predicted from sequence
            repseq_letter_annotations = g.protein.representative_sequence.seq_record.letter_annotations
            for k in repseq_letter_annotations.keys():
                if k != 'repchain_resnums':
                    to_append['{}_PRED'.format(k)] = repseq_letter_annotations[k][resnum - 1]

            if g.protein.representative_structure:
                if g.protein.representative_structure.is_experimental:
                    structure_type_suffix = 'EXP'
                else:
                    structure_type_suffix = 'HOM'

                # Structure properties - calculated
                # TODO: refactor to use reference_seq
                mapped = g.protein.representative_structure._map_repseq_resnums_to_repchain_index(resnum)
                if resnum in mapped:
                    repchain_index = mapped[resnum]

                    if not g.protein.representative_structure.representative_chain.seq_record:
                        appender.append(to_append)
                        continue

                    repchain_letter_annotations = g.protein.representative_structure.representative_chain.seq_record.letter_annotations
                    for k in repchain_letter_annotations.keys():
                        to_append['{}_{}'.format(k, structure_type_suffix)] = repchain_letter_annotations[k][repchain_index]

                    # At disulfide bond?
                    repchain_annotations = g.protein.representative_structure.representative_chain.seq_record.annotations
                    if 'SSBOND-biopython' in repchain_annotations:
                        structure_resnum = g.protein.representative_structure.map_repseq_resnums_to_structure_resnums(g.protein.representative_sequence,
                            resnum)
                        if resnum in structure_resnum:
                            ssbonds = repchain_annotations['SSBOND-biopython']
                            ssbonds_res = []
                            for x in ssbonds:
                                ssbonds_res.append(x[0])
                                ssbonds_res.append(x[1])

                            if structure_resnum in ssbonds_res:
                                to_append['at_disulfide_bridge'] = True

            appender.append(to_append)

        cols = ['base_residue', 'base_resnum', 'strain_residue', 'num_strains',
                'base_residue_prop', 'strain_residue_prop', 'grantham_score', 'grantham_annotation',
                'at_disulfide_bridge',
                'SS-sspro_PRED', 'SS-sspro8_PRED', 'SS-dssp_{}'.format(structure_type_suffix), 'TM-tmhmm_PRED',
                'RSA-accpro_PRED', 'RSA-accpro20_PRED', 'RSA-dssp_{}'.format(structure_type_suffix), 'ASA-dssp_{}'.format(structure_type_suffix),
                'CA_DEPTH-msms_{}'.format(structure_type_suffix), 'RES_DEPTH-msms_{}'.format(structure_type_suffix),
                'PHI-dssp_{}'.format(structure_type_suffix), 'PSI-dssp_{}'.format(structure_type_suffix), 'structure_resnums_{}'.format(structure_type_suffix),
                'strains_mutated']

        df_gene_summary = pd.DataFrame.from_records(appender, columns=cols)

        # Drop columns that don't have anything in them
        df_gene_summary.dropna(axis=1, how='all', inplace=True)

        df_gene_summary.sort_values(by='base_resnum', inplace=True)
        df_gene_summary = df_gene_summary.set_index('base_resnum')
        return df_gene_summary

    def download_mutation_images(self, outdir):
        # TODO: dunno if this works
        import ipywidgets
        import math

        views = []
        for g in self.base_strain_gempro.genes:
            if g.protein.representative_structure:
                view = g.protein.view_all_mutations(grouped=False, structure_opacity=0.5,
                                                    opacity_range=(0.6, 1), scale_range=(.5, 5))
                view._remote_call("setSize", target='Widget', args=['300px', '300px'])
                view.download_image(filename='{}_mutations.png'.format(g.id))
                views.append(view)

        hboxes = [ipywidgets.HBox(views[i * 3:i * 3 + 3])
                  for i in range(int(math.ceil(len(views) / 3.0)))]
        vbox = ipywidgets.VBox(hboxes)
        return vbox


from __future__ import print_function
import os
import sys
import datetime
import os.path as op
import glob
import pandas as pd
import numpy as np
import contextlib
import logging
import distutils.spawn
import subprocess
import shlex
import requests
import json
import warnings
import gzip
from collections import OrderedDict
from collections import Callable

log = logging.getLogger(__name__)


def is_ipynb():
    """Return True if the module is running in IPython kernel,
    False if in IPython shell or other Python shell.

    Copied from: http://stackoverflow.com/a/37661854/1592810
    There are other methods there too

    >>> is_ipynb()
    False

    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':  # Jupyter notebook or qtconsole?
            return True
        elif shell == 'TerminalInteractiveShell':  # Terminal running IPython?
            return False
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter


class Date():
    """Various methods to return formatted dates for today.
    """
    def __init__(self):
        self.today = datetime.date.today()
        self.short_date = self.today.strftime('%y%m%d')
        self.long_date = self.today.strftime('%Y-%m-%d')


def todays_short_date():
    today = Date()
    return today.short_date


def todays_long_date():
    today = Date()
    return today.long_date


class DefaultOrderedDict(OrderedDict):
    """Class to combine defaultdict and OrderedDict.
    Source: http://stackoverflow.com/a/6190500/562769

    """
    def __init__(self, default_factory=None, *a, **kw):
        if (default_factory is not None and
                not isinstance(default_factory, Callable)):
            raise TypeError('first argument must be callable')
        OrderedDict.__init__(self, *a, **kw)
        self.default_factory = default_factory

    def __getitem__(self, key):
        try:
            return OrderedDict.__getitem__(self, key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        self[key] = value = self.default_factory()
        return value

    def __reduce__(self):
        if self.default_factory is None:
            args = tuple()
        else:
            args = self.default_factory,
        return type(self), args, None, None, self.items()

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        return type(self)(self.default_factory, self)

    def __deepcopy__(self, memo):
        import copy
        return type(self)(self.default_factory,
                          copy.deepcopy(self.items()))

    def __repr__(self):
        return 'OrderedDefaultDict(%s, %s)' % (self.default_factory,
                                               OrderedDict.__repr__(self))


@contextlib.contextmanager
def suppress_stdout():
    """Suppress the stdout of any function.

    Usage:
    with ssbio.utils.suppress_stdout():
        my_function_here()
    """
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout


def clean_df(df, fill_nan=True, drop_empty_columns=True):
    """Clean a pandas dataframe by:
        1. Filling empty values with Nan
        2. Dropping columns with all empty values

    Args:
        df: Pandas DataFrame
        fill_nan (bool): If any empty values (strings, None, etc) should be replaced with NaN
        drop_empty_columns (bool): If columns whose values are all empty should be dropped

    Returns:
        DataFrame: cleaned DataFrame

    """
    if fill_nan:
        df = df.fillna(value=np.nan)
    if drop_empty_columns:
        df = df.dropna(axis=1, how='all')
    return df


def deprecated(func):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emmitted
    when the function is used."""
    def newFunc(*args, **kwargs):
        warnings.warn("Call to deprecated function %s." % func.__name__,
                      category=DeprecationWarning)
        return func(*args, **kwargs)
    newFunc.__name__ = func.__name__
    newFunc.__doc__ = func.__doc__
    newFunc.__dict__.update(func.__dict__)
    return newFunc


def split_folder_and_path(filepath):
    """Split a file path into its folder, filename, and extension

    Args:
        path (str): Path to a file

    Returns:
        tuple: of (folder, filename (without extension), extension)

    """
    dirname = op.dirname(filepath)
    filename = op.basename(filepath)
    splitext = op.splitext(filename)
    filename_without_extension = splitext[0]
    extension = splitext[1]

    return dirname, filename_without_extension, extension


def outfile_maker(inname, outext='.out', outname='', outdir='', append_to_name=''):
    """Create a default name for an output file based on the inname name, unless a output name is specified.

    Args:
        inname: Path to input file
        outext: Optional specified extension for output file (with the "."). Default is ".out".
        outfile: Optional specified name of output file.
        outdir: Optional path to output directory.

    Returns:
        str: Path to final output destination.

    Examples:

        >>> outfile_maker(inname='P00001.fasta')
        'P00001.out'

        >>> outfile_maker(inname='P00001')
        'P00001.out'

        >>> outfile_maker(inname='P00001.fasta', append_to_name='_new')
        'P00001_new.out'

        >>> outfile_maker(inname='P00001.fasta', outext='.mao')
        'P00001.mao'

        >>> outfile_maker(inname='P00001.fasta', outext='.mao', append_to_name='_new')
        'P00001_new.mao'

        >>> outfile_maker(inname='P00001.fasta', outext='.new', outname='P00001_aligned')
        'P00001_aligned.new'

        >>> outfile_maker(inname='P00001.fasta', outname='P00001_aligned')
        'P00001_aligned.out'

        >>> outfile_maker(inname='P00001.fasta', outname='P00001_aligned', append_to_name='_new')
        'P00001_aligned_new.out'

        >>> outfile_maker(inname='P00001.fasta', outname='P00001_aligned', outdir='/my/dir/')
        '/my/dir/P00001_aligned.out'

        >>> outfile_maker(inname='/test/other/dir/P00001.fasta', append_to_name='_new')
        '/test/other/dir/P00001_new.out'

        >>> outfile_maker(inname='/test/other/dir/P00001.fasta', outname='P00001_aligned')
        '/test/other/dir/P00001_aligned.out'

        >>> outfile_maker(inname='/test/other/dir/P00001.fasta', outname='P00001_aligned', outdir='/my/dir/')
        '/my/dir/P00001_aligned.out'

    """

    # TODO: CHECK IF OUTNAME IS A VALID FILE NAME!
    orig_dir, orig_name, orig_ext = split_folder_and_path(inname)

    # If output filename not provided, default is to take name of inname
    if not outname:
        outname = orig_name

    # Create new path in the same directory of old path if a new one isn't specified
    if not outdir:
        outdir = orig_dir

    # Append additional stuff to the filename if specified
    if append_to_name:
        outname += append_to_name

    # Join the output filename and output extension
    final_outfile = op.join(outdir, '{}{}'.format(outname, outext))

    return final_outfile


def force_rerun(flag, outfile):
    """Check if we should force rerunning of a command if an output file exists.

    Args:
        flag (bool): Flag to force rerun.
        outfile (str): Path to output file which may already exist.

    Returns:
        bool: If we should force rerunning of a command

    Examples:
        >>> force_rerun(flag=True, outfile='/not/existing/file.txt')
        True

        >>> force_rerun(flag=False, outfile='/not/existing/file.txt')
        True

        >>> force_rerun(flag=True, outfile='/existing/file.txt')
        True

        >>> force_rerun(flag=False, outfile='/existing/file.txt')
        False

    """
    # If flag is True, always run
    if flag:
        return True
    # If flag is False but file doesn't exist, also run
    elif not flag and not op.exists(outfile):
        return True
    # If flag is False but filesize of output is 0, also run
    elif not flag and not os.stat(outfile).st_size != 0:
        return True
    # Otherwise, do not run
    else:
        return False


def gunzip_file(infile, outfile=None, outdir=None, delete_original=False, force_rerun_flag=False):
    """Decompress a gzip file and optionally set output values.

    Args:
        infile: Path to .gz file
        outfile: Name of output file
        outdir: Path to output directory
        delete_original: If original .gz file should be deleted
        force_rerun_flag: If file should be decompressed if outfile already exists

    Returns:
        str: Path to decompressed file

    """
    if not outfile:
        outfile = infile.replace('.gz', '')

    if not outdir:
        outdir = ''
    else:
        outdir = op.dirname(infile)
    outfile = op.join(outdir, op.basename(outfile))

    if force_rerun(flag=force_rerun_flag, outfile=outfile):
        gz = gzip.open(infile, "rb")
        decoded = gz.read()

        with open(outfile, "wb") as new_file:
            new_file.write(decoded)

        gz.close()
        log.debug('{}: file unzipped'.format(outfile))
    else:
        log.debug('{}: file already unzipped'.format(outfile))

    if delete_original:
        os.remove(infile)

    return outfile


def request_file(link, outfile, force_rerun_flag=False):
    """Download a file given a URL if the outfile does not exist already.

    Args:
        link (str): Link to download file.
        outfile (str): Path to output file, will make a new file if it does not exist. Will not download if it does
            exist, unless force_rerun_flag is True.
        force_rerun_flag (bool): Flag to force re-downloading of the file if it exists already.

    Returns:
        str: Path to downloaded file.

    """
    if force_rerun(flag=force_rerun_flag, outfile=outfile):
        req = requests.get(link)
        if req.status_code == 200:
            with open(outfile, 'w') as f:
                f.write(req.text)
            log.debug('Loaded and saved {} to {}'.format(link, outfile))
        else:
            log.error('{}: request error {}'.format(link, req.status_code))
    return outfile


def request_json(link, outfile, outdir=None, force_rerun_flag=False):
    """Download a file in JSON format from a web request

    Args:
        link: Link to web request
        outfile: Name of output file
        outdir: Directory of output file
        force_rerun_flag: If true, redownload the file

    Returns:
        dict: contents of the JSON request

    """
    if not outdir:
        outdir = ''
    outfile = op.join(outdir, outfile)

    if force_rerun(flag=force_rerun_flag, outfile=outfile):
        text_raw = requests.get(link)
        my_dict = text_raw.json()
        with open(outfile, 'w') as f:
            json.dump(my_dict, f)

        log.debug('Loaded and saved {} to {}'.format(link, outfile))
    else:
        with open(outfile, 'r') as f:
            my_dict = json.load(f)
        log.debug('Loaded {}'.format(outfile))

    return my_dict

def program_exists(prog_name):
    """Check if a program is available as a command line executable on a system.

    Args:
        prog_name: Name of the program.

    Returns:
        bool: True if the program is available.

    """
    if distutils.spawn.find_executable(prog_name):
        return True
    else:
        return False


def command_runner(shell_command, force_rerun_flag, outfile, silent=False):
    """Run a program with command-line arguments.

    Args:
        program: Name of the program.
        args: Any program flags as they would be formatted in the command-line (ie. "-i test.in -o test.out").
        force_rerun_flag: If the program should be rerun even if the output file exists.
        outfile: Name out the output file which may have been generated.
            This does not specify what the outfile will be, that should be done in the args.

    Returns:
        bool: If the program ran successfully.

    """
    program_and_args = shlex.split(shell_command)

    # Check if program is installed
    if not program_exists(program_and_args[0]):
        raise OSError('{}: program not installed'.format(program_and_args[0]))

    # Check for force rerunning
    if force_rerun(flag=force_rerun_flag, outfile=outfile):
        if silent:
            command = subprocess.Popen(program_and_args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = command.communicate()
            ret = command.returncode
        else:
            # Prints output
            for path in execute(program_and_args):
                print(path, end="")

        # TODO: check return code and log properly
        log.debug('{}: Ran program, output to {}'.format(program_and_args[0], outfile))
    else:
        log.debug('{}: Output already exists'.format(outfile))


def execute(cmd):
    popen = subprocess.Popen(cmd, stdout=subprocess.PIPE, universal_newlines=True)
    for stdout_line in iter(popen.stdout.readline, ""):
        yield stdout_line
    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, cmd)


def write_torque_script(command, outfile, walltime, queue, name, out, err, print_exec=True):

    with open(outfile, 'w') as script:
        script.write('#PBS -l walltime={}\n'.format(walltime))
        script.write('#PBS -q regular\n')
        script.write('#PBS -N {}\n'.format(name))
        script.write('#PBS -o {}.out\n'.format(out))
        script.write('#PBS -e {}.err\n'.format(err))
        script.write('cd ${PBS_O_WORKDIR}\n')
        script.write(command)

    os.chmod(outfile, 0o755)

    if print_exec:
        print('qsub {};'.format(outfile))

    return outfile


def dict_head(d, N=5):
    """Return the head of a dictionary. It will be random!

    Default is to return the first 5 key/value pairs in a dictionary.

    Args:
        d: Dictionary to get head.
        N: Number of elements to display.

    Returns:
        dict: the first N items of the dictionary.

    """
    return {k: d[k] for k in list(d.keys())[:N]}


def rank_dated_files(pattern, dir, descending=True):
    """Search a directory for files that match a pattern. Return an ordered list of these files by filename.

    Args:
        pattern: The glob pattern to search for.
        dir: Path to directory where the files will be searched for.
        descending: Default True, will sort alphabetically by descending order.

    Returns:
        list: Rank-ordered list by filename.

    """
    files = glob.glob(op.join(dir, pattern))
    return sorted(files, reverse=descending)


def find(lst, a, case_sensitive=True):
    """Return indices of a list which have elements that match an object or list of objects

    Args:
        lst: list of values
        a: object(s) to check equality
        case_sensitive: if the search should be case sensitive

    Returns:
        list: list of indicies of lst which equal a

    """
    a = force_list(a)

    if not case_sensitive:
        lst = [x.lower() for x in lst]
        a = [y.lower() for y in a]

    return [i for i, x in enumerate(lst) if x in a]


def not_find(lst, a, case_sensitive=True):
    """Return indices of a list which have elements that DO NOT match an object or list of objects

    Args:
        lst: list of values
        a: object(s) to check inequality
        case_sensitive: if the search should be case sensitive

    Returns:
        list: list of indices of lst which do not equal a

    """
    a = force_list(a)

    if not case_sensitive:
        lst = [x.lower() for x in lst]
        a = [y.lower() for y in a]

    return [i for i, x in enumerate(lst) if x not in a]


def filter_list(lst, takeout, case_sensitive=True):
    """Return a modified list removing items specified.

    Args:
        lst: Original list of values
        takeout: Object or objects to remove from lst
        case_sensitive: if the search should be case sensitive

    Returns:
        list: Filtered list of values

    """
    takeout = force_list(takeout)

    if not case_sensitive:
        lst = [x.lower() for x in lst]
        takeout = [y.lower() for y in takeout]

    return [x for x in lst if x not in takeout]


def filter_list_by_indices(lst, indices):
    """Return a modified list containing only the indices indicated.

    Args:
        lst: Original list of values
        indices: List of indices to keep from the original list

    Returns:
        list: Filtered list of values

    """
    return [x for i, x in enumerate(lst) if i in indices]


def force_string(val=None):
    """Force a string representation of an object

    Args:
        val: object to parse into a string

    Returns:
        str: String representation

    """
    if val is None:
        return ''
    if isinstance(val, list):
        newval = [str(x) for x in val]
        return ';'.join(newval)
    if isinstance(val, str):
        return val
    else:
        return str(val)


def force_list(val=None):
    """Force a list representation of an object

    Args:
        val: object to parse into a list

    Returns:

    """
    if val is None:
        return []
    if isinstance(val, pd.Series):
        return val.tolist()
    return val if isinstance(val, list) else [val]


def force_lower_list(val=None):
    """Force a lowercase list representation of strings

    Args:
        val: string or strings to parse into a list

    Returns:
        list: with lowercase values

    """
    return [x.lower() for x in force_list(val)]


def force_upper_list(val=None):
    """Force a UPPERCASE list representation of strings

    Args:
        val: string or strings to parse into a list

    Returns:
        list: with UPPERCASE values

    """
    return [x.upper() for x in force_list(val)]


def split_list_by_n(l, n):
    """Split a list into lists of size n.

    Args:
        l: List of stuff.
        n: Size of new lists.

    Returns:
        list: List of lists each of size n derived from l.

    """
    n = max(1, n)
    return list(l[i:i+n] for i in range(0, len(l), n))


def split_list_into_n_lists(l, n):
    """Split a list into n lists.

    Args:
        l: List of stuff.
        n: Number of new lists to generate.

    Returns:
        list: List of n lists.

    """
    return [l[i::n] for i in range(n)]


def input_list_parser(infile_list):
    """Always return a list of files with varying input.

    >>> input_list_parser(['/path/to/folder/'])
    ['/path/to/folder/file1.txt', '/path/to/folder/file2.txt', '/path/to/folder/file3.txt']

    >>> input_list_parser(['/path/to/file.txt'])
    ['/path/to/file.txt']

    >>> input_list_parser(['file1.txt'])
    ['file1.txt']

    Args:
        infile_list: List of arguments

    Returns:
        list: Standardized list of files

    """

    final_list_of_files = []

    for x in infile_list:

        # If the input is a folder
        if op.isdir(x):
            os.chdir(x)
            final_list_of_files.extend(glob.glob('*'))

        # If the input is a file
        if op.isfile(x):
            final_list_of_files.append(x)

    return final_list_of_files


def flatlist_dropdup(list_of_lists):
    """Make a single list out of a list of lists, and drop all duplicates.

    Args:
        list_of_lists: List of lists.

    Returns:
        list: List of single objects.

    """
    return list(set([str(item) for sublist in list_of_lists for item in sublist]))


def combinations(iterable, r):
    """Calculate combinations

    >>> list(combinations('ABCD',2))
    [['A', 'B'], ['A', 'C'], ['A', 'D'], ['B', 'C'], ['B', 'D'], ['C', 'D']]

    >>> list(combinations(range(4), 3))
    [[0, 1, 2], [0, 1, 3], [0, 2, 3], [1, 2, 3]]

    Args:
        iterable: Any iterable object.
        r: Size of combination.

    Yields:
        list: Combination of size r.

    """
    pool = tuple(iterable)
    n = len(pool)
    if r > n:
        return

    indices = list(range(r))
    yield list(pool[i] for i in indices)

    while True:
        for i in reversed(range(r)):
            if indices[i] != i + n - r:
                break
        else:
            return
        indices[i] += 1
        for j in range(i + 1, r):
            indices[j] = indices[j - 1] + 1
        yield list(pool[i] for i in indices)


def percentage_to_float(x):
    """Convert a string representation of a percentage to float.

    >>> percentage_to_float('55%')
    0.55

    Args:
        x: String representation of a percentage

    Returns:
        float: Percentage in decimal form

    """
    return float(x.strip('%')) / 100


def make_dir(path):
    """Make a directory if it does not already exist

    Args:
        path: Path to new directory

    """
    if not op.exists(path):
        os.mkdir(path)
    return path


def remap( x, oMin, oMax, nMin, nMax ):
    """Map to a 0 to 1 scale
        http://stackoverflow.com/questions/929103/convert-a-number-range-to-another-range-maintaining-ratio

    """

    #range check
    if oMin == oMax:
        log.warning("Zero input range, unable to rescale")
        return x

    if nMin == nMax:
        log.warning("Zero output range, unable to rescale")
        return x

    #check reversed input range
    reverseInput = False
    oldMin = min( oMin, oMax )
    oldMax = max( oMin, oMax )
    if not oldMin == oMin:
        reverseInput = True

    #check reversed output range
    reverseOutput = False
    newMin = min( nMin, nMax )
    newMax = max( nMin, nMax )
    if not newMin == nMin :
        reverseOutput = True

    portion = (x-oldMin)*(newMax-newMin)/(oldMax-oldMin)
    if reverseInput:
        portion = (oldMax-x)*(newMax-newMin)/(oldMax-oldMin)

    result = portion + newMin
    if reverseOutput:
        result = newMax - portion

    return result


def scale_calculator(multiplier, elements, rescale=None):
    """Get a dictionary of scales for each element in elements.

    Examples:
        >>> scale_calculator(1, [2,7,8])
        {8: 1, 2: 1, 7: 1}

        >>> scale_calculator(1, [2,2,2,3,4,5,5,6,7,8])
        {2: 3, 3: 1, 4: 1, 5: 2, 6: 1, 7: 1, 8: 1}

        >>> scale_calculator(1, [2,2,2,3,4,5,5,6,7,8], rescale=(0.5,1))
        {2: 1.0, 3: 0.5, 4: 0.5, 5: 0.75, 6: 0.5, 7: 0.5, 8: 0.5}

        >>> scale_calculator(1, {2:3, 3:1, 4:1, 5:2, 6:1, 7:1, 8:1}, rescale=(0.5,1))
        {2: 1.0, 3: 0.5, 4: 0.5, 5: 0.75, 6: 0.5, 7: 0.5, 8: 0.5}

        >>> scale_calculator(1, [(2,2,2),(3,),(4,),(5,),(5,),(6,7,8)], rescale=(0.5,1))
        {(2, 2, 2): 0.5, (5,): 1.0, (3,): 0.5, (6, 7, 8): 0.5, (4,): 0.5}

    Args:
        mutiplier (int, float): Base float to be multiplied
        elements (list, dict): Dictionary which contains object:count
            or list of objects that may have repeats which will be counted
        rescale (tuple): Min and max values to rescale to

    Returns:
        dict: Scaled values of mutiplier for each element in elements

    """

    # TODO: think about what happens when:
    # TODO: 1. there is only one (or n) of each element, and rescale is set to seomthing. what is the original min/max to scale from?
    # TODO: 2. can we normalize the scale based on other counts? (ie. other gene mutation frequencies)

    if isinstance(elements, list):
        unique_elements = list(set(elements))
        scales = {}
        for x in unique_elements:
            count = elements.count(x)
            scales[x] = multiplier * count
    elif isinstance(elements, dict):
        scales = {}
        for k,count in elements.items():
            scales[k] = multiplier * int(count)
    else:
        raise ValueError('Input list of elements or dictionary of elements & counts')

    if not rescale:
        return scales
    else:
        new_scales = {}
        for k,v in scales.items():
            new_scales[k] = remap(v, min(scales.values()), max(scales.values()), rescale[0], rescale[1])
        return new_scales


from ssbio.core.object import Object
import logging
log = logging.getLogger(__name__)


class ChainProp(Object):
    """Class for protein structural properties of a specific chain"""

    def __init__(self, ident, pdb_parent, seq_record=None, description=None):
        Object.__init__(self, id=ident, description=description)

        self.pdb_parent = pdb_parent
        self.seq_record = seq_record

        if not self.description:
            self.description = 'Chain {} from PDB parent {}'.format(self.id, self.pdb_parent)

    def reset_seq_record(self):
        self.seq_record = None


from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.Alphabet import IUPAC


def cast_to_str(obj):
    """Return a string representation of a Seq or SeqRecord.

    Args:
        obj (str, Seq, SeqRecord): Biopython Seq or SeqRecord

    Returns:
        str: String representation of the sequence

    """

    if isinstance(obj, str):
        return obj
    if isinstance(obj, Seq):
        return str(obj)
    if isinstance(obj, SeqRecord):
        return str(obj.seq)
    else:
        raise ValueError('Must provide a string, Seq, or SeqRecord object.')


def cast_to_seq_record(obj, alphabet=IUPAC.extended_protein, id="<unknown id>", name="<unknown name>",
                       description="<unknown description>", dbxrefs=None,
                       features=None, annotations=None,
                       letter_annotations=None):
    """Return a SeqRecord representation of a string or Seq object.

    Args:
        obj (str, Seq, SeqRecord): Sequence string or Biopython Seq object
        alphabet: See Biopython SeqRecord docs
        id: See Biopython SeqRecord docs
        name: See Biopython SeqRecord docs
        description: See Biopython SeqRecord docs
        dbxrefs: See Biopython SeqRecord docs
        features: See Biopython SeqRecord docs
        annotations: See Biopython SeqRecord docs
        letter_annotations: See Biopython SeqRecord docs

    Returns:
        SeqRecord: SeqRecord representation of the sequence

    """

    if isinstance(obj, SeqRecord):
        return obj
    if isinstance(obj, Seq):
        return SeqRecord(obj, id, name, description, dbxrefs, features, annotations, letter_annotations)
    if isinstance(obj, str):
        obj = obj.upper()
        return SeqRecord(Seq(obj, alphabet), id, name, description, dbxrefs, features, annotations, letter_annotations)
    else:
        raise ValueError('Must provide a string, Seq, or SeqRecord object.')


import tempfile

from Bio import SeqIO

import ssbio.protein.sequence.utils
import ssbio.utils


def write_fasta_file(seq_records, outname, outdir=None, outext='.faa', force_rerun=False):
    """Write a FASTA file for a SeqRecord or a list of SeqRecord objects.

    Args:
        seq_records (SeqRecord, list): SeqRecord or a list of SeqRecord objects
        outname: Name of the output file which will have outext appended to it
        outdir: Path to directory to output sequences to
        outext: Extension of FASTA file, default ".faa"
        force_rerun: If file should be overwritten if it exists

    Returns:
        str: Path to output FASTA file.

    """

    if not outdir:
        outdir = ''
    outfile = ssbio.utils.outfile_maker(inname='', outname=outname, outdir=outdir, outext=outext)

    if ssbio.utils.force_rerun(flag=force_rerun, outfile=outfile):
        SeqIO.write(seq_records, outfile, "fasta")

    return outfile


def write_fasta_file_from_dict(indict, outname, outdir=None, outext='.faa', force_rerun=False):
    """Write a FASTA file for a dictionary of IDs and their sequence strings.

    Args:
        indict: Input dictionary with keys as IDs and values as sequence strings
        outname: Name of the output file which will have outext appended to it
        outdir: Path to directory to output sequences to
        outext: Extension of FASTA file, default ".faa"
        force_rerun: If file should be overwritten if it exists

    Returns:
        str: Path to output FASTA file.

    """

    if not outdir:
        outdir = ''
    outfile = ssbio.utils.outfile_maker(inname='', outname=outname, outdir=outdir, outext=outext)

    if ssbio.utils.force_rerun(flag=force_rerun, outfile=outfile):
        seqs = []
        for i, s in indict.items():
            seq = ssbio.protein.sequence.utils.cast_to_seq_record(s, id=i)
            seqs.append(seq)
        SeqIO.write(seqs, outfile, "fasta")

    return outfile


def write_seq_as_temp_fasta(seq):
    """Write a sequence as a temporary FASTA file

    Args:
        seq (str, Seq, SeqRecord): Sequence string, Biopython Seq or SeqRecord object

    Returns:
        str: Path to temporary FASTA file (located in system temporary files directory)

    """
    sr = ssbio.protein.sequence.utils.cast_to_seq_record(seq, id='tempfasta')
    return write_fasta_file(seq_records=sr, outname='temp', outdir=tempfile.gettempdir(), force_rerun=True)


def load_fasta_file(filename):
    """Load a FASTA file and return the sequences as a list of SeqRecords

    Args:
        filename (str): Path to the FASTA file to load

    Returns:
        list: list of all sequences in the FASTA file as Biopython SeqRecord objects

    """

    with open(filename, "r") as handle:
        records = list(SeqIO.parse(handle, "fasta"))
    return records


def load_fasta_file_as_dict_of_seqs(filename):
    """Load a FASTA file and return the sequences as a dict of {ID: sequence string}

    Args:
        filename (str): Path to the FASTA file to load

    Returns:
        dict: Dictionary of IDs to their sequence strings

    """

    results = {}
    records = load_fasta_file(filename)
    for r in records:
        results[r.id] = str(r.seq)

    return results


def fasta_files_equal(seq_file1, seq_file2):
    """Check equality of a FASTA file to another FASTA file

    Args:
        seq_file1: Path to a FASTA file
        seq_file2: Path to another FASTA file

    Returns:
        bool: If the sequences are the same

    """

    # Load already set representative sequence
    seq1 = SeqIO.read(open(seq_file1), 'fasta')

    # Load kegg sequence
    seq2 = SeqIO.read(open(seq_file2), 'fasta')

    # Test equality
    if str(seq1.seq) == str(seq2.seq):
        return True
    else:
        return False


import glob
import logging
import os

import numpy as np
import pandas as pd

import ssbio.protein.sequence.utils.alignment
from ssbio.utils import percentage_to_float

log = logging.getLogger(__name__)


def sequence_checker(reference_seq_aln, structure_seq_aln,
                     seq_ident_cutoff=0.5, allow_missing_on_termini=0.2,
                     allow_mutants=False, allow_deletions=False,
                     allow_insertions=False, allow_unresolved=False):
    """Report if a structure's sequence meets coverage checks to a reference sequence.

    First aligns a sequence from a chain of a PDB structure to "reference" sequence.
    Then creates a DataFrame of results to check for everything.

    Args:
        reference_seq_aln (str, Seq, SeqRecord): Reference sequence, alignment form
        structure_seq_aln (str, Seq, SeqRecord): Structure sequence, alignment form
        seq_ident_cutoff (float): Percent sequence identity cutoff, in decimal form
        allow_missing_on_termini (float): Percentage of the total length of the reference sequence which will be ignored
            when checking for modifications. Example: if 0.1, and reference sequence is 100 AA, then only residues
            5 to 95 will be checked for modifications.
        allow_mutants (bool): If mutations should be allowed or checked for
        allow_deletions (bool): If deletions should be allowed or checked for
        allow_insertions (bool): If insertions should be allowed or checked for
        allow_unresolved (bool): If unresolved residues should be allowed or checked for

    Returns:
        bool: If the structure's sequence meets the quality checks.

    """
    if len(reference_seq_aln) != len(structure_seq_aln):
        raise ValueError('Sequence lengths not equal - was an alignment run?')

    reference_seq_aln = ssbio.protein.sequence.utils.cast_to_str(reference_seq_aln)
    structure_seq_aln = ssbio.protein.sequence.utils.cast_to_str(structure_seq_aln)

    # Check percent identity cutoff
    stats_percent_ident = ssbio.protein.sequence.utils.alignment.get_percent_identity(reference_seq_aln, structure_seq_aln)
    log.debug('{}: percent identity'.format(stats_percent_ident))
    if stats_percent_ident < seq_ident_cutoff:
        log.debug('Alignment does not meet percent identity cutoff')
        return False
    else:
        log.debug('Alignment meets percent identity cutoff')

    # Parse the alignment results
    aln_df = ssbio.protein.sequence.utils.alignment.get_alignment_df(a_aln_seq=reference_seq_aln,
                                                                     b_aln_seq=structure_seq_aln)

    # Get cutoff stuff ready
    ref_seq_len = len(reference_seq_aln.replace('-', ''))
    allow_missing_on_termini /= 2
    # If any differences appear before start, they are ignored
    start = ref_seq_len - (ref_seq_len * (1 - allow_missing_on_termini))
    # If any differences appear before end, they are ignored
    end = ref_seq_len - (ref_seq_len * allow_missing_on_termini)

    no_deletions_in_pdb = False
    no_insertions_in_pdb = False
    no_mutants_in_pdb = False
    no_unresolved_in_pdb = False

    # Check everything
    if not allow_deletions:
        # Get indices of the deletions
        deletions = ssbio.protein.sequence.utils.alignment.get_deletions(aln_df)
        # If there are no deletions, that's great
        if len(deletions) == 0:
            log.debug('No deletion regions')
            no_deletions_in_pdb = True
        else:
            log.debug('{} deletion region(s)'.format(len(deletions)))
            # If the deletion appears before or after the cutoff, that's also great
            for deletion in deletions:
                if deletion[0][1] < start or deletion[0][0] > end:
                    no_deletions_in_pdb = True
                    log.debug('Deletion region(s) are not within structure core')
                else:
                    no_deletions_in_pdb = False
                    log.debug('Deletions within structure')
                    log.debug('{} < {} or {} > {}'.format(deletion[0][1], start, deletion[0][0], end))
                    break
    else:
        no_deletions_in_pdb = True

    if not allow_insertions:
        # Get indices of the insertions
        insertions = ssbio.protein.sequence.utils.alignment.get_insertions(aln_df)
        # If there are no insertions, that's great
        if len(insertions) == 0:
            log.debug('No insertion regions')
            no_insertions_in_pdb = True
        else:
            log.debug('{} insertion region(s)'.format(len(insertions)))
            # If the insertion appears before or after the cutoff, that's also great
            for insertion in insertions:
                if insertion[0][1] < start or insertion[0][0] > end:
                    no_insertions_in_pdb = True
                    log.debug('Insertion region(s) are not within structure core')
                else:
                    no_insertions_in_pdb = False
                    log.debug('Insertion regions within structure')
                    break
    else:
        no_insertions_in_pdb = True

    if not allow_mutants:
        # Get indices of the mutants
        mutations_full = ssbio.protein.sequence.utils.alignment.get_mutations(aln_df)
        mutations = [x[1] for x in mutations_full]
        # If there are no mutants, that's great
        if len(mutations) == 0:
            log.debug('No point mutations')
            no_mutants_in_pdb = True
        else:
            log.debug('{} point mutation(s)'.format(len(mutations)))
            # If the mutant appears before or after the cutoff, that's also great
            for mutation in mutations:
                if mutation < start or mutation > end:
                    no_mutants_in_pdb = True
                    log.debug('Mutation region(s) are not within structure core')
                else:
                    no_mutants_in_pdb = False
                    log.debug('Mutantion regions within structure')
                    break
    else:
        no_mutants_in_pdb = True

    if not allow_unresolved:
        # Get indices of the unresolved residues
        unresolved = ssbio.protein.sequence.utils.alignment.get_unresolved(aln_df)
        # If there are no unresolved, that's great
        if len(unresolved) == 0:
            log.debug('No unresolved mutations')
            no_unresolved_in_pdb = True
        else:
            log.debug('{} unresolved residue(s)'.format(len(unresolved)))
            # If the unresolved residue appears before or after the cutoff, that's also great
            for unr in unresolved:
                if unr < start or unr > end:
                    no_unresolved_in_pdb = True
                    log.debug('Unresolved region(s) are not within structure core')
                else:
                    no_unresolved_in_pdb = False
                    log.debug('Unresolved residues within structure')
                    break
    else:
        no_unresolved_in_pdb = True

    if no_deletions_in_pdb and no_insertions_in_pdb and no_mutants_in_pdb and no_unresolved_in_pdb:
        return True
    else:
        return False


def parse_procheck(quality_directory):
    """Parses all PROCHECK files in a directory and returns a Pandas DataFrame of the results

    Args:
        quality_directory: path to directory with PROCHECK output (.sum files)

    Returns:
        Pandas DataFrame: Summary of PROCHECK results

    """

    # TODO: save as dict instead, offer df as option
    # TODO: parse for one file instead

    procheck_summaries = glob.glob(os.path.join(quality_directory, '*.sum'))

    if len(procheck_summaries) == 0:
        return pd.DataFrame()

    all_procheck = {}
    for summ in procheck_summaries:
        structure_id = os.path.basename(summ).split('.sum')[0]
        procheck_dict = {}
        with open(summ) as f_in:
            lines = (line.rstrip() for line in f_in) # All lines including the blank ones
            lines = (line for line in lines if line) # Non-blank lines
            for line in lines:
                if len(line.split()) > 1:
                    if line.split()[1] == 'Ramachandran':
                        procheck_dict['procheck_rama_favored'] = percentage_to_float(line.split()[3])
                        procheck_dict['procheck_rama_allowed'] = percentage_to_float(line.split()[5])
                        procheck_dict['procheck_rama_allowed_plus'] = percentage_to_float(line.split()[7])
                        procheck_dict['procheck_rama_disallowed'] = percentage_to_float(line.split()[9])
                    if line.split()[1] == 'G-factors':
                        procheck_dict['procheck_gfac_dihedrals'] = line.split()[3]
                        procheck_dict['procheck_gfac_covalent'] = line.split()[5]
                        procheck_dict['procheck_gfac_overall'] = line.split()[7]
        all_procheck[structure_id] = procheck_dict

    DF_PROCHECK = pd.DataFrame.from_dict(all_procheck, orient='index')

    return DF_PROCHECK


def parse_psqs(psqs_results_file):
    """Parse a PSQS result file and returns a Pandas DataFrame of the results

    Args:
        psqs_results_file: Path to psqs results file

    Returns:
        Pandas DataFrame: Summary of PSQS results

    """

    # TODO: generalize column names for all results, save as dict instead

    psqs_results = pd.read_csv(psqs_results_file, sep='\t', header=None)
    psqs_results['pdb_file'] = psqs_results[0].apply(lambda x: str(x).strip('./').strip('.pdb'))
    psqs_results = psqs_results.rename(columns = {1:'psqs_local', 2:'psqs_burial', 3:'psqs_contact', 4:'psqs_total'}).drop(0, axis=1)
    psqs_results['u_pdb'] = psqs_results['pdb_file'].apply(lambda x: x.upper() if len(x)==4 else np.nan)
    psqs_results['i_entry_name'] = psqs_results['pdb_file'].apply(lambda x: x.split('_model1')[0] if len(x)>4 else np.nan)
    psqs_results = psqs_results[pd.notnull(psqs_results.psqs_total)]

    return psqs_results



import shlex
from Bio.PDB.MMCIF2Dict import MMCIF2Dict


class MMCIF2DictFix(MMCIF2Dict):
    """Fixes for MMCIF2Dict according to biopython#481 and biopython#523. Methods override parent.
    """

    def __init__(self, filename):
        super(MMCIF2DictFix, self).__init__(filename=filename)

    def _tokenize(self, handle):
        for line in handle:
            if line.startswith("#"):
                continue
            elif line.startswith(";"):
                token = line[1:].strip()
                for line in handle:
                    line = line.strip()
                    if line == ';':
                        break
                    token += line
                yield token
            else:
                try:
                    tokens = shlex.split(line)
                except ValueError:
                    # error "No closing quotation"
                    line = line.replace("'",'"')
                    # if odd - add a closing " to that line
                    if not line.count('"') % 2 == 0:
                        line = '{}"'.format(line)
                    tokens = shlex.split(line)
                for token in tokens:
                    yield token


import unittest


class Test(unittest.TestCase):
    """Unit tests for seqtools."""

    def test_load_fasta_file(self):
        pass

    def test_load_seq_string(self):
        pass

    def test_write_fasta_file(self):
        pass

if __name__ == "__main__":
    unittest.main()



import unittest

from ssbio.core.object import Object

class TestObject(unittest.TestCase):
    """Unit tests for Object"""

    @classmethod
    def setUpClass(self):
        self.ob = Object(id='idtester', description='nametester')

    def test_attr(self):
        self.assertTrue(hasattr(self.ob, 'id'))
        self.assertTrue(hasattr(self.ob, 'description'))

    def test_update(self):
        new_stuff = {'newkey':'newvalue', 'dontadd':'dontadd', 'description':'newdescription'}
        self.ob.update(newdata=new_stuff, overwrite=False, only_keys=['newkey','description'])
        self.assertTrue(hasattr(self.ob, 'newkey'))
        self.assertEqual('nametester', self.ob.description)

        self.ob.update(newdata=new_stuff, overwrite=True, only_keys=['description'])
        self.assertEqual('newdescription', self.ob.description)

        self.ob.update(newdata=new_stuff, overwrite=True)
        self.assertEqual('newdescription', self.ob.description)
        self.assertEqual('newvalue', self.ob.newkey)
        self.assertEqual('dontadd', self.ob.dontadd)

    def test_get_dict(self):
        gotdict = self.ob.get_dict(only_keys='id')
        self.assertEqual(gotdict, {'id':'idtester'})


from __future__ import print_function

import getpass
import logging
import os
import os.path as op

from ssbio.protein.sequence.utils import fasta as fasta

log = logging.getLogger(__name__)

class ITASSERPrep():
    """Prepare a sequence for I-TASSER runs
    """

    def __init__(self, ident, seq_str, root_dir, itasser_path, itlib_path,
                 execute_dir=None, light='true', runtype='local', print_exec=False,
                 binding_site_pred=False, ec_pred=False, go_pred=False, project_id=None,
                 slurm_email='', slurm_username='', walltime='48:00:00',
                 slurm_queue='shared'):
        """Create the I-TASSER folder and also an executable script to run I-TASSER

        Args:
            ident: Identifier for your sequence. Will be used as the global ID (folder name, sequence name)
            seq_str: Sequence in string format
            root_dir: Local directory where I-TASSER folder will be created
            itasser_path: Path to I-TASSER folder, i.e. '~/software/I-TASSER4.4'
            itlib_path: Path to ITLIB folder, i.e. '~/software/ITLIB'
            execute_dir: Optional path to execution directory - use this if you are copying the homology models to
                another location such as a supercomputer for running
            light: If simulations should be limited to 5 runs
            runtype: How you will be running I-TASSER - local, slurm, or torque
            print_exec: If the execution script should be printed out
            binding_site_pred: If binding site predictions should be run
            ec_pred: If EC number predictions should be run
            go_pred: If GO term predictions should be run
            walltime: SLURM or Torque walltime
            project_id: NERSC specific - SLURM project ID
            slurm_email: NERSC specific - SLURM email
            slurm_username: NERSC specific - SLURM username
            slurm_queue: NERSC specific - SLURM queue to use
        """

        if runtype.lower() not in ['local', 'torque', 'slurm']:
            raise ValueError('Invalid runtype, must be "local", "torque", "slurm"')

        self.id = ident
        self.seq_str = seq_str

        if not self.seq_str:
            raise ValueError('{}: no sequence input'.format(self.id))

        if len(self.seq_str) < 10 or len(self.seq_str) > 1500:
            log.warning('{}: I-TASSER modeling will not run as sequence length ({}) is not in the range [10, 1500]'.format(self.id, len(self.seq_str)))

        self.root_dir = root_dir
        if not op.exists(root_dir):
            os.makedirs(root_dir)
        if not execute_dir:
            # If no execute_dir is given, use the same dir as the created folder
            self.execute_dir = self.prep_folder(seq_str)
        elif execute_dir:
            orig_data_dir = self.prep_folder(seq_str)
            self.execute_dir = op.join(execute_dir, op.basename(orig_data_dir))

        self.print_exec = print_exec
        self.runtype = runtype
        self.light = light

        self.model_exists = op.exists(op.join(self.execute_dir, 'model1.pdb'))

        additional_options = ''
        if binding_site_pred:
            additional_options += '-LBS true '
        if ec_pred:
            additional_options += '-EC true '
        if go_pred:
            additional_options += '-GO true '
        self.additional_options = additional_options

        if runtype == 'local' or runtype == 'torque':
            self.walltime = walltime

            self.prep_script_local(itasser_loc=itasser_path,
                                   itlib_loc=itlib_path)

        if runtype == 'slurm':
            self.slurm_email = slurm_email
            self.slurm_username = slurm_username
            self.walltime = walltime
            self.slurm_queue = slurm_queue
            self.project_id = project_id

            self.prep_script_slurm(itasser_loc=itasser_path,
                                   itlib_loc=itlib_path)

    def prep_folder(self, seq):
        """Take in a sequence string and prepares the folder for the I-TASSER run
        """
        itasser_dir = op.join(self.root_dir, self.id)

        if not op.exists(itasser_dir):
            os.makedirs(itasser_dir)

        tmp = {self.id: seq}

        fasta.write_fasta_file_from_dict(indict=tmp,
                                         outname='seq',
                                         outext='.fasta',
                                         outdir=itasser_dir)
        return itasser_dir

    def prep_script_local(self, itasser_loc, itlib_loc, java_home='/usr/'):
        script_file = '{}.sh'.format(self.id)
        outfile = os.path.join(self.root_dir, script_file)

        itasser = {'executable': op.join(itasser_loc, 'I-TASSERmod/runI-TASSER.pl'),
                   'pkgdir': itasser_loc,
                   'libdir': itlib_loc,
                   'seqname': self.id,
                   'datadir': self.execute_dir,
                   'usrname': getpass.getuser(),
                   'java_home': java_home,
                   'additional_options': self.additional_options,
                   'walltime': self.walltime,
                   'light': self.light}

        script = open(outfile, 'w')
        script.write('#!/bin/bash -l\n')

        if self.runtype == 'torque':
            script.write('#PBS -l walltime={i[walltime]}\n'.format(i=itasser))
            script.write('#PBS -q regular\n')
            script.write('#PBS -N {i[seqname]}\n'.format(i=itasser))
            script.write('#PBS -o {i[seqname]}.out\n'.format(i=itasser))
            script.write('#PBS -e {i[seqname]}.err\n'.format(i=itasser))

        script.write(("{i[executable]} "
                      "-pkgdir {i[pkgdir]} "
                      "-libdir {i[libdir]} "
                      "-seqname {i[seqname]} "
                      "-datadir {i[datadir]} "
                      "-usrname {i[usrname]} "
                      "-java_home {i[java_home]} "
                      "{i[additional_options]}"
                      "-light {i[light]}\n\n").format(i=itasser))
        script.close()

        os.chmod(outfile, 0o755)

        if self.print_exec and self.runtype=='local':
            print('nohup ./{} > {}.out &'.format(op.basename(outfile), os.path.join(self.root_dir, self.id)),
                  end='\n\n')

        if self.print_exec and self.runtype == 'torque':
            print('qsub {}'.format(op.basename(outfile), os.path.join(self.root_dir, self.id)),
                  end='; ')

        return outfile

    def prep_script_slurm(self,
                          itasser_loc,
                          itlib_loc,
                          java_home='${JAVA_HOME}'
                          ):
        script_file = '{}.slm'.format(self.id)
        outfile = os.path.join(self.root_dir, script_file)

        itasser = {'executable': op.join(itasser_loc, 'I-TASSERmod/runI-TASSER.pl'),
                   'pkgdir': itasser_loc,
                   'libdir': itlib_loc,
                   'seqname': self.id,
                   'datadir': self.execute_dir,
                   'usrname': self.slurm_username,
                   'java_home': java_home,
                   'light': self.light,
                   'additional_options': self.additional_options,
                   'project_id':self.project_id,
                   'walltime': self.walltime,
                   'slurm_queue': self.slurm_queue,
                   'slurm_email': self.slurm_email}

        slurm = open(outfile, 'w')

        slurm.write('#!/bin/bash -l\n')
        slurm.write('#SBATCH -p shared\n')
        slurm.write('#SBATCH -A {i[project_id]}\n'.format(i=itasser))
        slurm.write('#SBATCH -t {i[walltime]}\n'.format(i=itasser))
        slurm.write('#SBATCH --mem=8GB\n')
        slurm.write('#SBATCH -J {i[seqname]}\n'.format(i=itasser))
        slurm.write('#SBATCH --mail-user {i[slurm_email]}\n'.format(i=itasser))
        slurm.write('#SBATCH -o {i[seqname]}.out\n'.format(i=itasser))
        slurm.write('#SBATCH -e {i[seqname]}.err\n'.format(i=itasser))
        slurm.write('#SBATCH --mail-type=BEGIN\n')
        slurm.write('#SBATCH --mail-type=END\n')
        slurm.write(('{i[executable]} '
                     '-pkgdir {i[pkgdir]} '
                     '-libdir {i[libdir]} '
                     '-seqname {i[seqname]} '
                     '-datadir {i[datadir]} '
                     '-usrname {i[usrname]} '
                     '-java_home {i[java_home]} '
                     '{i[additional_options]}'
                     '-light {i[light]}\n\n').format(i=itasser))

        slurm.close()

        os.chmod(outfile, 0o755)

        if self.print_exec:
            print('sbatch {}'.format(op.basename(outfile)), end='; ')

        return outfile


if __name__ == '__main__':
    pass

    # TODO: make this an executable script to
    # 1) ask for global I-TASSER locations
    # 2) ask for working directory
    # 3) take in multiple inputs and prepare them for I-TASSER runs
    #     a) input types
    #         i) a single FASTA file with single or multiple sequences
    #         ii) multiple FASTA files contained in the working directory
    #         iii) a dataframe with IDs and sequences
    #         iv) a sequence string and an ID (and optional additional identifiers)
    #     b) types of runs
    #         i) NERSC slurm (sbatch) inputs
    #         ii) local torque (qsub) inputs
    #         iii) simple executable background scripts
    # 4) Output executable scripts or submit things to the queue

    # root = '/home/nathan/projects/GEM-PRO/cyano/'
    # files = glob.glob(os.path.join(root,'*.faa'))
    # for f in files:
    #     identifier = os.path.splitext(os.path.basename(f))[0]
    #     ip = ITASSERPrep(id=identifier, root_dir='/home/nathan/projects/GEM-PRO/cyano')
    #
    #     sequence = sl.seq_loader(f, is_file=True)
    #     execute_dir = ip.prep_folder(sequence)
    #     ip.prep_script_local(itasser_loc='/home/nathan/software/I-TASSER4.4',
    #                          itlib_loc='/home/nathan/software/ITLIB',
    #                          datadir=execute_dir)

    # ip = ITASSERPrep(id='W5EP13', root_dir='/home/nathan/projects/GEM-PRO/cyano/')
    #
    # sequence = sl.seq_loader('/home/nathan/Downloads/W5EP13.faa', is_file=True)
    # execute_dir = ip.prep_folder(sequence)
    # ip.prep_script_local(itasser_loc='/home/nathan/software/I-TASSER4.4',
    #                      itlib_loc='/home/nathan/software/ITLIB',
    #                      datadir=execute_dir)


## below is old run_all script in python
# import os
# import shutil
# import subprocess
#
# thedir = '.'
# folders = [name for name in os.listdir(
#     thedir) if os.path.isdir(os.path.join(thedir, name))]
# folders = sorted(folders, reverse=True)
# for_ssb3 = folders[:len(folders) / 2]
#
# for fo in for_ssb3:
#     coach = open('%s_coach.sh' % fo, 'w')
#
#     coach.write('#!/bin/bash\n')
#     coach.write('#PBS -l walltime=05:20:00\n')
#     coach.write('#PBS -q regular\n')
#     coach.write('#PBS -N %s\n' % fo)
#     coach.write('perl ~/software/I-TASSER4.4/I-TASSERmod/runCOACH.pl -pkgdir /home/nathan/software/I-TASSER4.4 -libdir /home/nathan/software/ITLIB -protname %s -model model1.pdb -datadir /home/nathan/projects/GEM-PRO/yome/all_test/%s -GO true\n\n' % (fo, fo))
#
#     coach.close()
#
#     # subprocess.call('qsub %s_coach.sh;' % (fo), shell=True)
#     print('qsub %s_coach.sh;' % (fo)),


import argparse
import logging
import os
import os.path as op
import textwrap

from Bio import PDB
from tqdm import tqdm

import ssbio.utils
from ssbio.protein.structure.utils.structureio import StructureIO

log = logging.getLogger(__name__)


class CleanPDB(PDB.Select):
    """Selection rules to clean a PDB file

    These rules aim to:
    - Add missing chains to a PDB file
    - Select a single chain if noted
    - Remove alternate atom locations
    - Add atom occupancies
    - Add B (temperature) factors (default Biopython behavior)
    """

    def __init__(self, remove_atom_alt=True, remove_atom_hydrogen=True, keep_atom_alt_id='A', add_atom_occ=True,
                 remove_res_hetero=True, add_chain_id_if_empty='X', keep_chains=None):
        """Initialize the parameters which indicate what cleaning will occur

        Args:
            remove_atom_alt: Remove alternate positions
            remove_atom_hydrogen: Remove hydrogen atoms
            keep_atom_alt_id: If removing alternate positions, which alternate ID to keep
            add_atom_occ: Add atom occupancy fields if not present
            remove_res_hetero: Remove all HETATMs
            add_chain_id_if_empty: Add a chain ID if not present
            keep_chains: Keep only these chains
        """
        self.remove_atom_alt = remove_atom_alt
        self.remove_atom_hydrogen = remove_atom_hydrogen
        self.keep_atom_alt_id = keep_atom_alt_id
        self.add_atom_occ = add_atom_occ
        self.remove_res_hetero = remove_res_hetero
        self.add_chain_id_if_empty = add_chain_id_if_empty
        if not keep_chains:
            self.keep_chains = []
        else:
            self.keep_chains = ssbio.utils.force_list(keep_chains)

    def accept_chain(self, chain):
        # If the chain does not have an ID, add one to it and keep it
        # http://comments.gmane.org/gmane.comp.python.bio.devel/10639
        if self.add_chain_id_if_empty and not chain.id.strip():
            chain.id = self.add_chain_id_if_empty
            return True
        # If a chain is specified and the current chain equals that specified chain, keep it
        elif self.keep_chains and chain.id in self.keep_chains:
            return True
        # If a chain is specified but the current chain does not equal that specified chain, remove it
        elif self.keep_chains and chain.id not in self.keep_chains:
            return False
        # If no chain is specified, keep all chains
        else:
            return True

    def accept_residue(self, residue):
        hetfield, resseq, icode = residue.get_id()
        if hetfield == '':
            hetfield = ' '
        # If you want to remove residues that are not normal, remove them
        if self.remove_res_hetero and hetfield[0] != ' ':
            return False
        else:
            return True

    def accept_atom(self, atom):
        # If the you want to remove hydrogens and the atom is a H, remove it
        if self.remove_atom_hydrogen and atom.element == 'H':
            return False
        # If you want to remove alternate locations, and the alt location isn't the one you want to keep, remove it
        elif self.remove_atom_alt and atom.is_disordered() and atom.get_altloc() != self.keep_atom_alt_id:
            return False
        else:
            # Add occupancies if there are none and you want to
            # http://comments.gmane.org/gmane.comp.python.bio.general/6289
            if self.add_atom_occ and atom.occupancy is None:
                atom.set_occupancy(1)
            if self.remove_atom_alt:
                atom.set_altloc(' ')
            return True

def clean_pdb(pdb_file, out_suffix='_clean', outdir=None, force_rerun=False,
              remove_atom_alt=True, remove_atom_hydrogen=True, keep_atom_alt_id='A', add_atom_occ=True,
              remove_res_hetero=True, add_chain_id_if_empty='X', keep_chains=None):
    """Clean a PDB file.

    Args:
        pdb_file: Path to input PDB file
        out_suffix: Suffix to append to original filename
        outdir: Path to output directory
        force_rerun: If structure should be re-cleaned if a clean file exists already
        remove_atom_alt: Remove alternate positions
        remove_atom_hydrogen: Remove hydrogen atoms
        keep_atom_alt_id: If removing alternate positions, which alternate ID to keep
        add_atom_occ: Add atom occupancy fields if not present
        remove_res_hetero: Remove all HETATMs
        add_chain_id_if_empty: Add a chain ID if not present
        keep_chains: Keep only these chains

    Returns:
        str: Path to cleaned PDB file

    """
    outfile = ssbio.utils.outfile_maker(inname=pdb_file,
                                        append_to_name=out_suffix,
                                        outdir=outdir,
                                        outext='.pdb')

    if ssbio.utils.force_rerun(flag=force_rerun, outfile=outfile):
        my_pdb = StructureIO(pdb_file)
        my_cleaner = CleanPDB(remove_atom_alt=remove_atom_alt,
                              remove_atom_hydrogen=remove_atom_hydrogen,
                              keep_atom_alt_id=keep_atom_alt_id,
                              add_atom_occ=add_atom_occ,
                              remove_res_hetero=remove_res_hetero,
                              add_chain_id_if_empty=add_chain_id_if_empty,
                              keep_chains=keep_chains)

        my_clean_pdb = my_pdb.write_pdb(out_suffix=out_suffix,
                                        out_dir=outdir,
                                        custom_selection=my_cleaner,
                                        force_rerun=force_rerun)

        return my_clean_pdb
    else:
        return outfile


if __name__ == '__main__':
    p = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                description=textwrap.dedent("""\
                                Clean PDB files - cleanpdb.py
                                -----------------------------
                                This script will automatically:

                                * Add missing chains to a PDB file
                                * Select a single chain or chains if noted
                                * Remove alternate atom locations
                                * Add atom occupancies
                                * Add B (temperature) factors (default Biopython behavior)

                                Cleaned PDBs will be in a clean_pdbs folder where the script is executed.
                                Example: script help
                                $ cleanpdb --help

                                Example: clean one PDB file
                                $ cleanpdb 1kf6.pdb

                                Example: clean one PDB file and keep only chains A and B
                                $ cleanpdb 1kf6.pdb --chain A,B

                                Example: clean multiple PDB files
                                $ cleanpdb *.pdb

                                Example: clean a whole directory of PDB
                                $ cleanpdb /path/to/pdb/files
                                """))
    p.add_argument('infile', help='PDB file or folder you want to clean', nargs='+', type=str)
    p.add_argument('--outsuffix', '-os', default='_clean', help='Suffix appended to PDB file')
    p.add_argument('--outdir', '-od', default='clean_pdbs', help='Directory to output clean PDBs')
    p.add_argument('--chain', '-c', default=None, help='Keep only specified chains')
    p.add_argument('--keephydro', '-hy', action='store_false', help='Keep hydrogen atoms (default is to remove)')
    p.add_argument('--keephetero', '-ht', action='store_false', help='Keep hetero atoms (default is to remove)')
    # TODO: if this flag is present, the alternate positions seem to switch line positions
    p.add_argument('--keepalt', '-ka', action='store_false', help='Keep alternate positions (default is to remove)')
    p.add_argument('--force', '-f', action='store_true', help='Force rerunning of cleaning even if the clean PDB exists')
    args = p.parse_args()

    if args.chain:
        args.chain = args.chain.split(',')

    if not op.isdir(args.outdir):
        os.mkdir(args.outdir)

    infiles = ssbio.utils.input_list_parser(args.infile)

    for pdb in tqdm(infiles):

        outfile = ssbio.utils.outfile_maker(inname=pdb,
                                            append_to_name=args.outsuffix,
                                            outdir=args.outdir,
                                            outext='.pdb')

        if ssbio.utils.force_rerun(flag=args.force, outfile=outfile):

            my_pdb = StructureIO(pdb)
            my_cleaner = CleanPDB(remove_atom_alt=args.keepalt,
                                  remove_atom_hydrogen=args.keephydro,
                                  keep_atom_alt_id='A',
                                  add_atom_occ=True,
                                  remove_res_hetero=args.keephetero,
                                  add_chain_id_if_empty='X',
                                  keep_chains=args.chain)

            my_clean_pdb = my_pdb.write_pdb(out_suffix=args.outsuffix,
                                            out_dir=args.outdir,
                                            custom_selection=my_cleaner,
                                            force_rerun=args.force)

    print('Clean PDBs at: {}'.format(args.outdir))



#!/usr/bin/env python

import os
import os.path as op
from pkg_resources import resource_filename


def prep_leap_folders(wd):
    if not op.exists(op.join(wd, 'xleap_modified')):
        os.mkdir(op.join(wd, 'xleap_modified'))
    if not op.exists(op.join(wd, 'amber_minimized')):
        os.mkdir(op.join(wd, 'amber_minimized'))

def make_run_all_script(wd):
    run_all = """#!/bin/bash

for i in *.pdb; do
    echo "Running ${i}..."
    seq=`echo ${i}  | cut -d. -f1`
    curr_dir=`pwd`

    cp ${i} temp.pdb

    tleap -f leaprc

    rm temp.pdb
    mv temp_mod.pdb xleap_modified/${seq}_xleap.pdb
    mv temp_mod.prmtop amber_minimized/${seq}.prmtop
    mv temp_mod.inpcrd amber_minimized/${seq}.inpcrd

done
"""
    my_file = op.join(wd, 'run_all.sh')
    with open(my_file, 'w') as f:
        f.write(run_all)
    os.chmod(my_file, 0o755)
    return my_file

def make_leaprc_file(wd):
    frcmod = resource_filename('ssbio.etc', 'frcmod.ff99SB')

    leaprc = """logFile leap.log
#
# ----- leaprc for loading the ff99SB (Hornak & Simmerling) force field
#
#	load atom type hybridizations
#
addAtomTypes {
	{ "H"   "H" "sp3" }
	{ "HO"  "H" "sp3" }
	{ "HS"  "H" "sp3" }
	{ "H1"  "H" "sp3" }
	{ "H2"  "H" "sp3" }
	{ "H3"  "H" "sp3" }
	{ "H4"  "H" "sp3" }
	{ "H5"  "H" "sp3" }
	{ "HW"  "H" "sp3" }
	{ "HC"  "H" "sp3" }
	{ "HA"  "H" "sp3" }
	{ "HP"  "H" "sp3" }
	{ "OH"  "O" "sp3" }
	{ "OS"  "O" "sp3" }
	{ "O"   "O" "sp2" }
	{ "O2"  "O" "sp2" }
	{ "OW"  "O" "sp3" }
	{ "CT"  "C" "sp3" }
	{ "CH"  "C" "sp3" }
	{ "C2"  "C" "sp3" }
	{ "C3"  "C" "sp3" }
	{ "C"   "C" "sp2" }
	{ "C*"  "C" "sp2" }
	{ "CA"  "C" "sp2" }
	{ "CB"  "C" "sp2" }
	{ "CC"  "C" "sp2" }
	{ "CN"  "C" "sp2" }
	{ "CM"  "C" "sp2" }
	{ "CK"  "C" "sp2" }
	{ "CQ"  "C" "sp2" }
	{ "CD"  "C" "sp2" }
	{ "CE"  "C" "sp2" }
	{ "CF"  "C" "sp2" }
	{ "CG"  "C" "sp2" }
	{ "CP"  "C" "sp2" }
	{ "CI"  "C" "sp2" }
	{ "CJ"  "C" "sp2" }
	{ "CW"  "C" "sp2" }
	{ "CV"  "C" "sp2" }
	{ "CR"  "C" "sp2" }
	{ "CA"  "C" "sp2" }
	{ "CY"  "C" "sp2" }
	{ "C0"  "C" "sp2" }
	{ "MG"  "Mg" "sp3" }
	{ "N"   "N" "sp2" }
	{ "NA"  "N" "sp2" }
	{ "N2"  "N" "sp2" }
	{ "N*"  "N" "sp2" }
	{ "NP"  "N" "sp2" }
	{ "NQ"  "N" "sp2" }
	{ "NB"  "N" "sp2" }
	{ "NC"  "N" "sp2" }
	{ "NT"  "N" "sp3" }
	{ "N3"  "N" "sp3" }
	{ "S"   "S" "sp3" }
	{ "SH"  "S" "sp3" }
	{ "P"   "P" "sp3" }
	{ "LP"  ""  "sp3" }
	{ "F"   "F" "sp3" }
	{ "CL"  "Cl" "sp3" }
	{ "BR"  "Br" "sp3" }
	{ "I"   "I"  "sp3" }
	{ "FE"  "Fe" "sp3" }
	{ "EP"  ""  "sp3" }
# glycam
	{ "OG"  "O" "sp3" }
	{ "OL"  "O" "sp3" }
	{ "AC"  "C" "sp3" }
	{ "EC"  "C" "sp3" }
}
#
#	Load the main parameter set.
#
parm99 = loadamberparams parm99.dat
frcmod99SB = loadamberparams %s
#
#	Load DNA/RNA libraries
#
loadOff all_nucleic94.lib
#
#	Load main chain and terminating
#	amino acid libraries (i.e. ff94 libs)
#
loadOff all_amino94.lib
loadOff all_aminoct94.lib
loadOff all_aminont94.lib
#
#       Load water and ions
#
loadOff ions94.lib
loadOff solvents.lib
HOH = TP3
WAT = TP3

#
#	Define the PDB name map for the amino acids and DNA.
#
addPdbResMap {
  { 0 "ALA" "NALA" } { 1 "ALA" "CALA" }
  { 0 "ARG" "NARG" } { 1 "ARG" "CARG" }
  { 0 "ASN" "NASN" } { 1 "ASN" "CASN" }
  { 0 "ASP" "NASP" } { 1 "ASP" "CASP" }
  { 0 "CYS" "NCYS" } { 1 "CYS" "CCYS" }
  { 0 "CYX" "NCYX" } { 1 "CYX" "CCYX" }
  { 0 "GLN" "NGLN" } { 1 "GLN" "CGLN" }
  { 0 "GLU" "NGLU" } { 1 "GLU" "CGLU" }
  { 0 "GLY" "NGLY" } { 1 "GLY" "CGLY" }
  { 0 "HID" "NHID" } { 1 "HID" "CHID" }
  { 0 "HIE" "NHIE" } { 1 "HIE" "CHIE" }
  { 0 "HIP" "NHIP" } { 1 "HIP" "CHIP" }
  { 0 "ILE" "NILE" } { 1 "ILE" "CILE" }
  { 0 "LEU" "NLEU" } { 1 "LEU" "CLEU" }
  { 0 "LYS" "NLYS" } { 1 "LYS" "CLYS" }
  { 0 "MET" "NMET" } { 1 "MET" "CMET" }
  { 0 "PHE" "NPHE" } { 1 "PHE" "CPHE" }
  { 0 "PRO" "NPRO" } { 1 "PRO" "CPRO" }
  { 0 "SER" "NSER" } { 1 "SER" "CSER" }
  { 0 "THR" "NTHR" } { 1 "THR" "CTHR" }
  { 0 "TRP" "NTRP" } { 1 "TRP" "CTRP" }
  { 0 "TYR" "NTYR" } { 1 "TYR" "CTYR" }
  { 0 "VAL" "NVAL" } { 1 "VAL" "CVAL" }
  { 0 "HIS" "NHIS" } { 1 "HIS" "CHIS" }
  { 0 "GUA" "DG5"  } { 1 "GUA" "DG3"  } { "GUA" "DG" }
  { 0 "ADE" "DA5"  } { 1 "ADE" "DA3"  } { "ADE" "DA" }
  { 0 "CYT" "DC5"  } { 1 "CYT" "DC3"  } { "CYT" "DC" }
  { 0 "THY" "DT5"  } { 1 "THY" "DT3"  } { "THY" "DT" }
  { 0 "G" "DG5"  } { 1 "G" "DG3"  } { "G" "DG" } { "GN" "DGN" }
  { 0 "A" "DA5"  } { 1 "A" "DA3"  } { "A" "DA" } { "AN" "DAN" }
  { 0 "C" "DC5"  } { 1 "C" "DC3"  } { "C" "DC" } { "CN" "DCN" }
  { 0 "T" "DT5"  } { 1 "T" "DT3"  } { "T" "DT" } { "TN" "DTN" }
  { 0 "C5" "DC5" }
  { 0 "G5" "DG5" }
  { 0 "A5" "DA5" }
  { 0 "T5" "DT5" }
  { 1 "C3" "DC3" }
  { 1 "G3" "DG3" }
  { 1 "A3" "DA3" }
  { 1 "T3" "DT3" }

}

addPdbAtomMap {
  { "O5*" "O5'" }
  { "C5*" "C5'" }
  { "C4*" "C4'" }
  { "O4*" "O4'" }
  { "C3*" "C3'" }
  { "O3*" "O3'" }
  { "C2*" "C2'" }
  { "C1*" "C1'" }
  { "C5M" "C7"  }
  { "H1*" "H1'" }
  { "H2*1" "H2'1" }
  { "H2*2" "H2'2" }
  { "H3*" "H3'" }
  { "H4*" "H4'" }
  { "H5*1" "H5'1" }
  { "H5*2" "H5'2" }
# old ff atom names -> new
  { "O1'" "O4'" }
  { "OA"  "O1P" }
  { "OB"  "O2P" }
}


#
# assumed that most often proteins use HIE
# Leap adds hydrogens based on which hydrogens are present in the library file for that particular 3-letter codon. For instance, HIE and HIS are the same and are singly-protonated at the epsilon position. HID is singly-protonated at the delta position, and HIP is doubly-protonated.
#
NHIS = NHIE
HIS = HIE
CHIS = CHIE


########################################


########################################
#########################################


wt = loadpdb temp.pdb
setbox wt vdw
#solvatebox wt TIP3PBOX 15
#addIons wt Na+ 0
deselect wt
relax wt
savepdb wt temp_mod.pdb
saveamberparm wt temp_mod.prmtop temp_mod.inpcrd
quit
""" % (frcmod)

    my_file = op.join(wd, 'leaprc')
    with open(my_file, 'w') as f:
        f.write(leaprc)
    return my_file

if __name__ == '__main__':
    """Run tleap for a specified directory of PDB files
    """
    import subprocess
    import argparse

    # TODO: move main function to standalone tleapauto script

    p = argparse.ArgumentParser(description='Run tleap on a folder of PDB files')
    p.add_argument('indir', help='Directory containing only PDB files which will run through leap')
    args = p.parse_args()

    # TODO: add prep md runs

    prep_leap_folders(args.indir)
    leaprc = make_leaprc_file(args.indir)
    run_all = make_run_all_script(args.indir)
    subprocess.call(run_all)



import gzip
import json
import logging
import os.path as op

import pandas as pd
import requests
import ssbio.protein.structure.complexes as cplx
from Bio.PDB.MMCIF2Dict import MMCIF2Dict
from lxml import etree
from six.moves.urllib.request import urlopen, urlretrieve

import ssbio.utils
from ssbio.protein.structure.structprop import StructProp

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

log = logging.getLogger(__name__)


class PDBProp(StructProp):
    """Class to parse through PDB properties
    """

    def __init__(self, ident, description=None, chains=None, mapped_chains=None, structure_path=None, file_type=None,
                 representative_chain=None):
        StructProp.__init__(self, ident, description=description, chains=chains, mapped_chains=mapped_chains,
                            structure_path=structure_path, file_type=file_type,
                            representative_chain=representative_chain, is_experimental=True)
        self.experimental_method = None
        self.resolution = None
        self.date = None
        self.taxonomy_name = None

    def download_structure_file(self, outdir, file_type, force_rerun=False):
        pdb_file = download_structure(pdb_id=self.id, file_type=file_type, only_header=False, outdir=outdir,
                                      force_rerun=force_rerun)
        log.debug('{}: downloaded {} file'.format(self.id, file_type))
        self.load_structure_path(pdb_file, file_type)

        if 'cif' in file_type:
            self.update(parse_mmcif_header(pdb_file))

    def download_cif_header_file(self, outdir, force_rerun=False):
        file_type = 'cif'
        cif_file = download_structure(pdb_id=self.id, file_type=file_type, only_header=True,
                                      outdir=outdir, force_rerun=force_rerun)
        log.debug('{}: downloaded mmCIF file header'.format(self.id))

        cif_dict = parse_mmcif_header(cif_file)
        self.update(cif_dict)

    def get_pisa_complex_predictions(self, outdir, existing_pisa_multimer_xml=None):
        if not self.is_experimental:
            log.error('Complex predictions for homology models currently not implemented')
            return

        if not existing_pisa_multimer_xml:
            pisa_xmls = cplx.download_pisa_multimers_xml(pdb_ids=self.id, outdir=outdir,
                                                         save_single_xml_files=True)
        else:
            pisa_xmls = {}
            pisa_xmls[self.id] = existing_pisa_multimer_xml
        pisa_dict = cplx.parse_pisa_multimers_xml(pisa_xmls[self.id], download_structures=True,
                                                  outdir=outdir)


def download_structure(pdb_id, file_type, outdir='', outfile='', only_header=False, force_rerun=False):
    """Download a structure from the RCSB PDB by ID. Specify the file type desired.

    Args:
        pdb_id: PDB ID
        file_type: pdb, pdb.gz, mmcif, cif, cif.gz, xml.gz, mmtf, mmtf.gz
        outdir: Optional output directory
        outfile: Optional output name
        only_header: If only the header file should be downloaded
        force_rerun: If the file should be downloaded again even if it exists

    Returns:
        str: Path to outfile

    """
    # TODO: keep an eye on https://github.com/biopython/biopython/pull/943 Biopython PR#493 for functionality of this
    # method in biopython. extra file types have not been added to biopython download yet

    pdb_id = pdb_id.lower()
    file_type = file_type.lower()
    file_types = ['pdb', 'pdb.gz', 'mmcif', 'cif', 'cif.gz', 'xml.gz', 'mmtf', 'mmtf.gz']
    if file_type not in file_types:
        raise ValueError('Invalid file type, must be either: pdb, pdb.gz, cif, cif.gz, xml.gz, mmtf, mmtf.gz')

    if file_type == 'mmtf':
        file_type = 'mmtf.gz'

    if file_type.endswith('.gz'):
        gzipped = True
    else:
        gzipped = False

    if file_type == 'mmcif':
        file_type = 'cif'

    if only_header:
        folder = 'header'
        if outfile:
            outfile = op.join(outdir, outfile)
        else:
            outfile = op.join(outdir, '{}.header.{}'.format(pdb_id, file_type))
    else:
        folder = 'download'
        if outfile:
            outfile = op.join(outdir, outfile)
        else:
            outfile = op.join(outdir, '{}.{}'.format(pdb_id, file_type))

    if ssbio.utils.force_rerun(flag=force_rerun, outfile=outfile):
        if file_type == 'mmtf.gz' or file_type == 'mmtf':
            mmtf_api = '1.0'
            download_link = 'http://mmtf.rcsb.org/v{}/full/{}.mmtf.gz'.format(mmtf_api, pdb_id)
        else:
            download_link = 'https://files.rcsb.org/{}/{}.{}'.format(folder, pdb_id, file_type)

        urlretrieve(download_link, outfile)

        if gzipped:
            outfile = ssbio.utils.gunzip_file(infile=outfile,
                                              outfile=outfile.strip('.gz'),
                                              outdir=outdir,
                                              delete_original=True,
                                              force_rerun_flag=force_rerun)

        log.debug('{}: saved structure file'.format(outfile))
    else:
        log.debug('{}: structure file already saved'.format(outfile))

    return outfile


def parse_pdb_header(infile):
    """Parse a couple important fields from the mmCIF file format with some manual curation of ligands.

    If you want full access to the mmCIF file just use the MMCIF2Dict class in Biopython.

    Args:
        infile: Path to mmCIF file

    Returns:
        dict: Dictionary of parsed header

    """
    pass


def parse_mmcif_header(infile):
    """Parse a couple important fields from the mmCIF file format with some manual curation of ligands.

    If you want full access to the mmCIF file just use the MMCIF2Dict class in Biopython.

    Args:
        infile: Path to mmCIF file

    Returns:
        dict: Dictionary of parsed header

    """
    newdict = {}
    try:
        mmdict = MMCIF2Dict(infile)
    except ValueError as e:
        log.exception(e)
        return newdict

    chemical_ids_exclude = ['HOH']
    chemical_types_exclude = ['l-peptide linking','peptide linking']

    if '_struct.title' in mmdict:
        newdict['pdb_title'] = mmdict['_struct.title']
    else:
        log.debug('{}: No title field'.format(infile))

    if '_struct.pdbx_descriptor' in mmdict:
        newdict['description'] = mmdict['_struct.pdbx_descriptor']
    else:
        log.debug('{}: no description field'),format(infile)

    if '_database_PDB_rev.date' in mmdict:
        newdict['date'] = mmdict['_database_PDB_rev.date']
    else:
        log.debug('{}: no date field'),format(infile)

    if '_exptl.method' in mmdict:
        newdict['experimental_method'] = mmdict['_exptl.method']
    else:
        log.debug('{}: no experimental method field'.format(infile))

    # TODO: refactor how to get resolutions based on experimental method
    if '_refine.ls_d_res_high' in mmdict:
        try:
            if isinstance(mmdict['_refine.ls_d_res_high'], list):
                newdict['resolution'] = [float(x) for x in mmdict['_refine.ls_d_res_high']]
            else:
                newdict['resolution'] = float(mmdict['_refine.ls_d_res_high'])
        except:
            newdict['resolution'] = float(mmdict['_em_3d_reconstruction.resolution'])
    else:
        log.debug('{}: no resolution field'.format(infile))

    if '_chem_comp.id' in mmdict:
        chemicals_filtered = ssbio.utils.filter_list_by_indices(mmdict['_chem_comp.id'],
                                                            ssbio.utils.not_find(mmdict['_chem_comp.type'],
                                                                           chemical_types_exclude,
                                                                           case_sensitive=False))
        chemicals_fitered = ssbio.utils.filter_list(chemicals_filtered, chemical_ids_exclude, case_sensitive=True)
        newdict['chemicals'] = chemicals_fitered
    else:
        log.debug('{}: no chemical composition field'.format(infile))

    if '_entity_src_gen.pdbx_gene_src_scientific_name' in mmdict:
        newdict['taxonomy_name'] = mmdict['_entity_src_gen.pdbx_gene_src_scientific_name']
    else:
        log.debug('{}: no organism field'.format(infile))

    return newdict


# TODO: check this for python 2
def download_sifts_xml(pdb_id, outdir='', outfile=''):
    """Download the SIFTS file for a PDB ID.

    Args:
        pdb_id:
        outdir:
        outfile:

    Returns:

    """
    baseURL = 'ftp://ftp.ebi.ac.uk/pub/databases/msd/sifts/xml/'
    filename = '{}.xml.gz'.format(pdb_id)

    if outfile:
        outfile = op.join(outdir, outfile)
    else:
        outfile = op.join(outdir, filename.split('.')[0] + '.sifts.xml')

    if not op.exists(outfile):
        response = urlopen(baseURL + filename)
        with open(outfile, 'wb') as f:
            f.write(gzip.decompress(response.read()))

    return outfile


def map_uniprot_resnum_to_pdb(uniprot_resnum, chain_id, sifts_file):
    """Map a UniProt residue number to its corresponding PDB residue number.

    This function requires that the SIFTS file be downloaded,
    and also a chain ID (as different chains may have different mappings).

    Args:
        uniprot_resnum (int): integer of the residue number you'd like to map
        chain_id (str): string of the PDB chain to map to
        sifts_file (str): Path to the SIFTS XML file

    Returns:
        (tuple): tuple containing:

            mapped_resnum (int): Mapped residue number
            is_observed (bool): Indicates if the 3D structure actually shows the residue

    """
    # Load the xml with lxml
    parser = etree.XMLParser(ns_clean=True)
    tree = etree.parse(sifts_file, parser)
    root = tree.getroot()

    my_pdb_resnum = None

    # TODO: "Engineered_Mutation is also a possible annotation, need to figure out what to do with that
    my_pdb_annotation = False

    # Find the right chain (entities in the xml doc)
    ent = './/{http://www.ebi.ac.uk/pdbe/docs/sifts/eFamily.xsd}entity'
    for chain in root.findall(ent):
        # TODO: IMPORTANT - entityId is not the chain ID!!! it is just in alphabetical order!
        if chain.attrib['entityId'] == chain_id:
            # Find the "crossRefDb" tag that has the attributes dbSource="UniProt" and  dbResNum="your_resnum_here"
            # Then match it to the crossRefDb dbResNum that has the attribute dbSource="PDBresnum"

            # Check if uniprot + resnum even exists in the sifts file (it won't if the pdb doesn't contain the residue)
            ures = './/{http://www.ebi.ac.uk/pdbe/docs/sifts/eFamily.xsd}crossRefDb[@dbSource="UniProt"][@dbResNum="%s"]' % uniprot_resnum
            my_uniprot_residue = chain.findall(ures)
            if len(my_uniprot_residue) == 1:
                # Get crossRefDb dbSource="PDB"
                parent = my_uniprot_residue[0].getparent()
                pres = './/{http://www.ebi.ac.uk/pdbe/docs/sifts/eFamily.xsd}crossRefDb[@dbSource="PDB"]'
                my_pdb_residue = parent.findall(pres)
                my_pdb_resnum = int(my_pdb_residue[0].attrib['dbResNum'])

                # Get <residueDetail dbSource="PDBe" property="Annotation">
                # Will be Not_Observed if it is not seen in the PDB
                anno = './/{http://www.ebi.ac.uk/pdbe/docs/sifts/eFamily.xsd}residueDetail[@dbSource="PDBe"][@property="Annotation"]'
                my_pdb_annotation = parent.findall(anno)
                if len(my_pdb_annotation) == 1:
                    my_pdb_annotation = my_pdb_annotation[0].text
                    if my_pdb_annotation == 'Not_Observed':
                        my_pdb_annotation = False
                else:
                    my_pdb_annotation = True
            else:
                return None, False

    return my_pdb_resnum, my_pdb_annotation


def best_structures(uniprot_id, outname=None, outdir=None, seq_ident_cutoff=0.0, force_rerun=False):
    """Use the PDBe REST service to query for the best PDB structures for a UniProt ID.

        More information found here: https://www.ebi.ac.uk/pdbe/api/doc/sifts.html
        Link used to retrieve results: https://www.ebi.ac.uk/pdbe/api/mappings/best_structures/:accession
        The list of PDB structures mapping to a UniProt accession sorted by coverage of the protein and, if the same, resolution.

        Here is the ranking algorithm described by the PDB paper:
        https://nar.oxfordjournals.org/content/44/D1/D385.full

        "Finally, a single quality indicator is also calculated for each entry by taking the harmonic average
        of all the percentile scores representing model and model-data-fit quality measures and then subtracting
        10 times the numerical value of the resolution (in Angstrom) of the entry to ensure that resolution plays
        a role in characterising the quality of a structure. This single empirical 'quality measure' value is used
        by the PDBe query system to sort results and identify the 'best' structure in a given context. At present,
        entries determined by methods other than X-ray crystallography do not have similar data quality information
        available and are not considered as 'best structures'."

        Args:
            uniprot_id (str): UniProt Accession ID
            outname (str): Basename of the output file of JSON results
            outdir (str): Path to output directory of JSON results
            seq_ident_cutoff (float): Cutoff results based on percent coverage (in decimal form)
            force_rerun (bool): Obtain best structures mapping ignoring previously downloaded results

        Returns:
            list: Rank-ordered list of dictionaries representing chain-specific PDB entries. Keys are:
                pdb_id: the PDB ID which maps to the UniProt ID
                chain_id: the specific chain of the PDB which maps to the UniProt ID
                coverage: the percent coverage of the entire UniProt sequence
                resolution: the resolution of the structure
                start: the structure residue number which maps to the start of the mapped sequence
                end: the structure residue number which maps to the end of the mapped sequence
                unp_start: the sequence residue number which maps to the structure start
                unp_end: the sequence residue number which maps to the structure end
                experimental_method: type of experiment used to determine structure
                tax_id: taxonomic ID of the protein's original organism

    """
    outfile = ''

    if not outdir:
        outdir = ''

    # if output dir is specified but not outname, use the uniprot
    if not outname and outdir:
        outname = uniprot_id

    if outname:
        outname = op.join(outdir, outname)
        outfile = '{}.json'.format(outname)

    # Load a possibly existing json file
    if not ssbio.utils.force_rerun(flag=force_rerun, outfile=outfile):
        with open(outfile, 'r') as f:
            raw_data = json.load(f)
        log.debug('{}: Loaded existing json file'.format(uniprot_id))

    # Otherwise run the web request
    else:
        # TODO: add a checker for a cached file of uniprot -> PDBs - can be generated within gempro pipeline and stored
        response = requests.get('https://www.ebi.ac.uk/pdbe/api/mappings/best_structures/{}'.format(uniprot_id),
                                data={'key': 'value'})
        if response.status_code == 404:
            log.debug('{}: 404 returned, probably no structures available.'.format(uniprot_id))
            raw_data = {uniprot_id: {}}
        else:
            log.debug('{}: Obtained best structures'.format(uniprot_id))
            raw_data = response.json()

        # Write the json file if specified
        if outfile:
            with open(outfile, 'w') as f:
                json.dump(raw_data, f)
            log.debug('{}: Saved json file of best structures'.format(uniprot_id))

    data = dict(raw_data)[uniprot_id]

    # Filter for sequence identity percentage
    if seq_ident_cutoff != 0:
        for result in data:
            if result['coverage'] < seq_ident_cutoff:
                data.remove(result)

    return data


def blast_pdb(seq, outfile='', outdir='', evalue=0.0001, seq_ident_cutoff=0.0, link=False, force_rerun=False):
    """Returns a list of BLAST hits of a sequence to available structures in the PDB.

    Args:
        seq (str): Your sequence, in string format
        outfile (str): Name of output file
        outdir (str, optional): Path to output directory. Default is the current directory.
        evalue (float, optional): Cutoff for the E-value - filters for significant hits. 0.001 is liberal, 0.0001 is stringent (default).
        seq_ident_cutoff (float, optional): Cutoff results based on percent coverage (in decimal form)
        link (bool, optional): Set to True if a link to the HTML results should be displayed
        force_rerun (bool, optional): If existing BLAST results should not be used, set to True. Default is False

    Returns:
        list: Rank ordered list of BLAST hits in dictionaries.

    """

    if len(seq) < 12:
        raise ValueError('Sequence must be at least 12 residues long.')
    if link:
        page = 'PDB results page: http://www.rcsb.org/pdb/rest/getBlastPDB1?sequence={}&eCutOff={}&maskLowComplexity=yes&matrix=BLOSUM62&outputFormat=HTML'.format(seq, evalue)
        print(page)

    parser = etree.XMLParser(ns_clean=True)

    outfile = op.join(outdir, outfile)
    if ssbio.utils.force_rerun(force_rerun, outfile):
        # Load the BLAST XML results if force_rerun=True
        page = 'http://www.rcsb.org/pdb/rest/getBlastPDB1?sequence={}&eCutOff={}&maskLowComplexity=yes&matrix=BLOSUM62&outputFormat=XML'.format(
                        seq, evalue)
        req = requests.get(page)
        if req.status_code == 200:
            response = req.text

            # Save the XML file
            if outfile:
                with open(outfile, 'w') as f:
                    f.write(response)

            # Parse the XML string
            tree = etree.ElementTree(etree.fromstring(response, parser))
            log.debug('Loaded BLAST results from REST server')
        else:
            log.error('BLAST request timed out')
            return []
    else:
        tree = etree.parse(outfile, parser)
        log.debug('{}: Loaded existing BLAST XML results'.format(outfile))

    # Get length of original sequence to calculate percentages
    len_orig = float(len(seq))

    root = tree.getroot()
    hit_list = []

    for hit in root.findall('BlastOutput_iterations/Iteration/Iteration_hits/Hit'):
        info = {}

        hitdef = hit.find('Hit_def')
        if hitdef is not None:
            info['hit_pdb'] = hitdef.text.split('|')[0].split(':')[0].lower()
            info['hit_pdb_chains'] = hitdef.text.split('|')[0].split(':')[2].split(',')

        # One PDB can align to different parts of the sequence
        # Will just choose the top hit for this single PDB
        hsp = hit.findall('Hit_hsps/Hsp')[0]

        # Number of identical residues
        hspi = hsp.find('Hsp_identity')
        if hspi is not None:
            info['hit_num_ident'] = int(hspi.text)
            info['hit_percent_ident'] = int(hspi.text)/len_orig

            if int(hspi.text)/len_orig < seq_ident_cutoff:
                log.debug('{}: does not meet sequence identity cutoff'.format(hitdef.text.split('|')[0].split(':')[0]))
                continue

        # Number of similar residues (positive hits)
        hspp = hsp.find('Hsp_positive')
        if hspp is not None:
            info['hit_num_similar'] = int(hspp.text)
            info['hit_percent_similar'] = int(hspp.text) / len_orig

        # Total number of gaps (unable to align in either query or subject)
        hspg = hsp.find('Hsp_gaps')
        if hspg is not None:
            info['hit_num_gaps'] = int(hspg.text)
            info['hit_percent_gaps'] = int(hspg.text) / len_orig

        # E-value of BLAST
        hspe = hsp.find('Hsp_evalue')
        if hspe is not None:
            info['hit_evalue'] = float(hspe.text)

        # Score of BLAST
        hsps = hsp.find('Hsp_score')
        if hsps is not None:
            info['hit_score'] = float(hsps.text)

        hit_list.append(info)

    log.debug("{}: Number of BLAST hits".format(len(hit_list)))
    return hit_list


def blast_pdb_df(blast_results):
    """Make a dataframe of BLAST results"""
    cols = ['hit_pdb', 'hit_pdb_chains', 'hit_evalue', 'hit_score', 'hit_num_ident', 'hit_percent_ident',
            'hit_num_similar', 'hit_percent_similar', 'hit_num_gaps', 'hit_percent_gaps']
    return pd.DataFrame.from_records(blast_results, columns=cols)


def _property_table():
    """Download the PDB -> resolution table directly from the RCSB PDB REST service.

    See the other fields that you can get here: http://www.rcsb.org/pdb/results/reportField.do

    Returns:
        Pandas DataFrame: table of structureId as the index, resolution and experimentalTechnique as the columns

    """
    url = 'http://www.rcsb.org/pdb/rest/customReport.csv?pdbids=*&customReportColumns=structureId,resolution,experimentalTechnique,releaseDate&service=wsfile&format=csv'
    r = requests.get(url)
    p = pd.read_csv(StringIO(r.text)).set_index('structureId')
    return p

def get_resolution(pdb_id):
    """Quick way to get the resolution of a PDB ID using the table of results from the REST service

    Returns infinity if the resolution is not available.

    Returns:
        float: resolution of a PDB ID in Angstroms

    TODO:
        - Unit test

    """

    pdb_id = pdb_id.upper()
    if pdb_id not in _property_table().index:
        raise ValueError('PDB ID not in property table')
    else:
        resolution = _property_table().ix[pdb_id, 'resolution']
        if pd.isnull(resolution):
            log.debug('{}: no resolution available, probably not an X-ray crystal structure')
            resolution = float('inf')

    return resolution


def get_release_date(pdb_id):
    """Quick way to get the release date of a PDB ID using the table of results from the REST service

    Returns None if the release date is not available.

    Returns:
        str: Organism of a PDB ID

    """

    pdb_id = pdb_id.upper()
    if pdb_id not in _property_table().index:
        raise ValueError('PDB ID not in property table')
    else:
        release_date = _property_table().ix[pdb_id, 'releaseDate']
        if pd.isnull(release_date):
            log.debug('{}: no taxonomy available')
            release_date = None

    return release_date


import unittest

import ssbio.protein.sequence.properties.thermostability


class TestThermostability(unittest.TestCase):
    """Unit tests for thermostability calculations
    """

    def test_gene_isos_to_uniprot(self):
        gene_1 = '52'
        isos_1 = ['1', '2', '3']
        uniprot_isos_1 = ['P24666-1', 'P24666-2', 'P24666-3']
        unrev_uniprot_isos_1 = ['A0A0B4J207-1']
        expected_1 = {'1': 'P24666-1', '2': 'P24666-2', '3': 'P24666-3'}
        self.assertEqual(
            ssbio.databases.isoform.gene_isos_to_uniprot(gene_1, isos_1, uniprot_isos_1, unrev_uniprot_isos_1),
            expected_1)


import os
import ssbio.utils
import pickle
from copy import deepcopy
import pandas as pd
import ssbio.utils
import ssbio.core.io
import logging
log = logging.getLogger(__name__)

class Object(object):
    """Cobra core object with additional methods to update and get attributes"""

    def __init__(self, id=None, description=None):
        self.id = id
        self.description = description

    def __str__(self):
        return str(self.id)

    def __repr__(self):
        return "<%s %s at 0x%x>" % (self.__class__.__name__, self.id, id(self))

    def update(self, newdata, overwrite=False, only_keys=None):
        """Add/update any attributes from a dictionary.

        Args:
            newdata: Dictionary of attributes
            overwrite: If existing attributes should be overwritten if provided in newdata
            only_keys: List of keys to update

        Examples:
            >>> myobj = Object(id='hi', description='blankname')
            >>> myobj.update({'description':'withname'}, overwrite=False)
            >>> myobj.get_dict() == {'id': 'hi', 'description':'blankname'}
            True

            >>> myobj = Object(id='hi', description='blankname')
            >>> myobj.update({'description':'withname'}, overwrite=True)
            >>> myobj.get_dict() == {'id': 'hi', 'description':'withname'}
            True

            >>> myobj = Object(id='hi', description='blankname')
            >>> myobj.update({'description':'withname', 'randomkey':'randomval'}, overwrite=True)
            >>> myobj.get_dict() == {'id': 'hi', 'description': 'withname', 'randomkey': 'randomval'}
            True

            >>> myobj = Object(id='hi', description='blankname')
            >>> myobj.update({'description':'withname', 'randomkey':'randomval'}, overwrite=True, only_keys='description')
            >>> myobj.get_dict() == {'id': 'hi', 'description': 'withname'}
            True

            >>> myobj = Object(id='hi', description='blankname')
            >>> myobj.update({'description':'withname', 'randomkey':'randomval'}, overwrite=True, only_keys='randomkey')
            >>> myobj.get_dict() == {'id': 'hi', 'description': 'blankname', 'randomkey': 'randomval'}
            True

            >>> myobj = Object(id='hi', description='blankname')
            >>> myobj.update({'description':'withname', 'randomkey':'randomval'}, overwrite=False)
            >>> myobj.get_dict() == {'id': 'hi', 'description': 'blankname', 'randomkey': 'randomval'}
            True

            >>> myobj = Object(id='hi', description='blankname')
            >>> myobj.update({'description':'withname', 'randomkey':'randomval'}, overwrite=False, only_keys='randomkey')
            >>> myobj.get_dict() == {'id': 'hi', 'description': 'blankname', 'randomkey': 'randomval'}
            True

        """
        # Filter for list of keys in only_keys
        if only_keys:
            only_keys = ssbio.utils.force_list(only_keys)
            newdata = {k:v for k,v in newdata.items() if k in only_keys}

        # Update attributes
        for key, value in newdata.items():
            # Overwrite flag overwrites all attributes
            if overwrite:
                setattr(self, key, value)
            else:
                # Otherwise check if attribute is None and set it if so
                if hasattr(self, key):
                    if not getattr(self, key):
                        setattr(self, key, value)
                    else:
                        continue
                # Or just add a new attribute
                else:
                    setattr(self, key, value)

    def get_dict(self, only_keys=None, exclude_attributes=None, df_format=False):
        """Get a dictionary of this object's attributes. Optional format for storage in a Pandas DataFrame.

        Args:
            keys (str, list): Attributes that should be returned. If not provided, all are returned.
            exclude_attributes (str, list): Attributes that should be excluded.
            df_format (bool): If dictionary values should be formatted for a dataframe
                (everything possible is transformed into strings, int, or float -
                if something can't be transformed it is excluded)

        Returns:
            dict: Dictionary of attributes

        """

        # Choose attributes to return, return everything in the object if a list is not specified
        if not only_keys:
            keys = list(self.__dict__.keys())
        else:
            keys = ssbio.utils.force_list(only_keys)

        # Remove keys you don't want returned
        if exclude_attributes:
            exclude_attributes = ssbio.utils.force_list(exclude_attributes)
            for x in exclude_attributes:
                if x in keys:
                    keys.remove(x)

        # Copy attributes into a new dictionary
        df_dict = {}
        for k, orig_v in self.__dict__.items():
            if k in keys:
                v = deepcopy(orig_v)
                if df_format:
                    if v and not isinstance(v, str) and not isinstance(v, int) and not isinstance(v, float) and not isinstance(v, bool):
                        try:
                            df_dict[k] = ssbio.utils.force_string(deepcopy(v))
                        except TypeError:
                            log.warning('{}: excluding attribute from dict, cannot transform into string'.format(k))
                    elif not v and not isinstance(v, int) and not isinstance(v, float):
                        df_dict[k] = None
                    else:
                        df_dict[k] = deepcopy(v)
                else:
                    df_dict[k] = deepcopy(v)
        return df_dict

    def save_dataframes(self, outdir, prefix='df_'):
        """Save all attributes that start with "df" into a specified directory.

        Args:
            outdir (str): Path to output directory
            prefix (str): Prefix that dataframe attributes start with

        """
        # Get list of attributes that start with "df_"
        dfs = list(filter(lambda x: x.startswith(prefix), dir(self)))

        counter = 0
        for df in dfs:
            outpath = ssbio.utils.outfile_maker(inname=df, outext='.csv', outdir=outdir)
            my_df = getattr(self, df)
            if not isinstance(my_df, pd.DataFrame):
                raise TypeError('{}: object is not a Pandas DataFrame'.format(df))

            if my_df.empty:
                log.debug('{}: empty dataframe, not saving'.format(df))
            else:
                my_df.to_csv(outpath)
                log.debug('{}: saved dataframe'.format(outpath))
                counter += 1

        log.debug('Saved {} dataframes at {}'.format(counter, outdir))

    def save_pickle(self, outfile, protocol=2):
        """Save the object as a pickle file

        Args:
            outfile (str): Filename
            protocol (int): Pickle protocol to use. Default is 2 to remain compatible with Python 2

        Returns:
            str: Path to pickle file

        """
        ssbio.core.io.save_pickle(self, outfile, protocol)

    def __json_encode__(self):
        to_return = {}
        # Don't save properties, methods in the JSON
        for x in [a for a in dir(self) if not a.startswith('__') and not a.startswith('_{}__'.format(type(self).__name__)) and not isinstance(getattr(type(self), a, None), property) and not callable(getattr(self,a))]:
            to_return.update({x: getattr(self, x)})
        return to_return

    def save_json(self, outfile, compression=True):
        """Save the object as a JSON file using json_tricks"""
        ssbio.core.io.save_json(self, outfile, compression=compression)


import os.path as op
import tempfile
import unittest

import ssbio.databases.pdb

class TestPDB(unittest.TestCase):
    """Unit tests for PDB
    """
    # # TODO: this result changes based on pdb database
    # def test_top_pdb_blast_hit(self):
    #     seq = 'VLSPADKTNVKAAWGVKALSPADKTNVKAALTAVAHVDDMPNAL'
    #     evalue = 1
    #     results = {'hit_evalue': '0.0640608',
    #                  'hit_num_ident': '36',
    #                  'hit_pdb': '3IA3',
    #                  'hit_pdb_chain': 'B,D',
    #                  'hit_percent_ident': 0.81818181818181823,
    #                  'hit_score': '76'}
    #     blast_hit = ssbio.databases.pdb.top_pdb_blast_hit(seq, evalue=evalue)
    #     for x in results.keys():
    #         self.assertTrue(x in list(blast_hit.keys()))